import abc

import numpy as np
from pmutt import constants as c

from ..Units import eval_qty
from ..Error import OutsideCorrelationError
from .. import yaml_io


class ThermochemBase(object):
    metaclass = abc.ABCMeta
    """Manage thermochemical properties correlation (abstract base class).

    This class cannot be instantiated directly; rather, it serves as a
    common base for the rest of the correlations.
    """
    # Abstract base class

    def init_params(self, range=None, units=None):
        if range is not None:
            self.range = (
                eval_qty(range[0], 'K').in_units('K'),
                eval_qty(range[1], 'K').in_units('K'))
        else:
            self.range = None

    def __init__(self, range=None, units=None):
        """
        Initialize generic thermochemical property correlation.

        Parameters
        ----------
        range : tuple(float, float), optional
            ``(lb, ub) = range`` where lb and ub are respectively the lower and
             uppers bounds of temperatures [K] for which the correlation is
             valid.
        """
        if range is not None:
            range = tuple(range)
            assert len(range) == 2
            assert range[1] >= range[0]
        self.range = range

    def check_range(self, T):
        """
        Verify that one or more temperatures is within the valid range for this
        correlation.

        Parameters
        ----------
        T : float or float array
            Specify temperature(s) to test.

        Returns
        -------
        None

        Raises
        ------
        chemtk.error.OutsideCorrelationError
            If `T` (or any temperature in array `T`) is outside the valid range
            of this correlation.
        """
        if self.range is None:
            return
        if np.any(T < self.range[0]) or np.any(T > self.range[1]):
            raise OutsideCorrelationError(
                "In %s().check_range: one or more of the following "
                "temperatures is outside the valid interval [%r,%r]: %r"
                % (type(self).__name__, self.range[0], self.range[1], T))

    def get_range(self):
        """
        Return range of validity for this thermochemical properties
        correlation.

        Returns
        -------
        lower_bound : float
            Lower bound of range.
        upper_bound : float
            Upper bound of range.
        """
        return self.range

    def set_range(self, range=None):
        """
        Set range of validity for this thermochemical properties correlation.

        Parameters
        ----------
        range : tuple(float, float), optional
            ``(lb, ub) = range`` where lb and ub are respectively the lower and
             uppers bounds of temperatures [K] for which the correlation is
             valid.
        """
        self.range = range

    @abc.abstractmethod
    def get_CpoR(self, T):
        """Return non-dimensional standard state heat capacity |eq_ND_Cp_T|."""

    @abc.abstractmethod
    def get_SoR(self, T, S_elements=None):
        """Return non-dimensional standard state entropy |eq_ND_S_T|."""

    @abc.abstractmethod
    def get_HoRT(self, T):
        """Return non-dimensional standard heat of formation |eq_ND_H_T|."""

    def get_GoRT(self, T, S_elements=None):
        """
        Return non-dimensional standard Gibbs energy of formation |eq_ND_G_T|.
        """
        return self.get_HoRT(T) - self.get_SoR(T, S_elements=S_elements)

    def get_H(self, T, units):
        """
        Return dimensional enthalpy at T with units specified by 'units'
        """
        return self.get_HoRT(T)*T*c.R('{}/K'.format(units))

    def get_G(self, T, units, S_elements=None):
        """
        Return dimensional Gibbs free energy at T with units
        specified by 'units'
        """
        return self.get_GoRT(T, S_elements=S_elements)*T*c.R('{}/K'.format(units))

    def get_S(self, T, units, S_elements=None):
        """
        Return dimensional entropy with units specified by 'units'
        """
        return self.get_SoR(T, S_elements=S_elements)*c.R(units)

    def get_Cp(self, T, units):
        """
        Return dimensional heat capacity with units specified by 'units'
        """
        return self.get_CpoR(T)*c.R(units)

    _yaml_schema = """
range:
  type: tuple
  item_types: [{type: qty, kind: temperature},
               {type: qty, kind: temperature}]
  optional: True
  desc: range of valid temperatures
"""


yaml_io.register_class('ThermochemBase',
                       yaml_io.parse(ThermochemBase._yaml_schema),
                       ThermochemBase)


__all__ = ['ThermochemBase']



"""
==================
Physical constants
==================

Values of common physical constants are defined here as instances of
:class:`pgradd.Units.Quantity`.

"""

from . Units import eval_qty


#: Planck's constant
PLANCK_CONSTANT = eval_qty('6.626068*10^-34 J s')

#: Boltzman's constant
BOLTZMANN_CONSTANT = eval_qty('1.3806503*10^-23 J/K')

#: Boltzman's constant
GAS_CONSTANT = eval_qty('8.314472 J/(mol K)')

#: Avogadro's number
AVOGADRO_NUMBER = eval_qty('6.02214179*10^23 mol^-1')



#  coding: utf-8
from .Parser import All                 # This parse class will go through them
#                                         until it runs out.
from .Parser import Either              # This parse class will require at
#                                         least one attribute inside to occur.
#                                         (other wise error)
from .Parser import Optional            # attributes of this class does
#                                         is optional
from .Parser import Literals            # literal string comaprison (several
#                                         string) with sort. So that Cl is
#                                         checked for C.
from .Parser import Digit               # 1 digit
from .Parser import Number              # any length number
from .Parser import String              # any length string
from .Parser import Filler              # Fillers are for checking if correct
#                                         syntanx are given. Are not recorded.
"""
This file contains Abstract Syntax Tree of the RING input. Parser grabs the
tree info from here, and interpret string.
strict_grammar follows the original RING paper abstract tree

Reference:

Rangarajan, S., Kaminski, T., Van Wyk, E., Bhan A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Algorithms of RING", Computers and Chemical Engineering 64 (2014) 124,
    10.1016/j.compchemeng.2014.02.007
Rangarajan, S., Bhan, A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Descrpition of RING", Computers and Chemical Engineering 45 (2012) 114,
    10.1016/j.compchemeng.2012.06.008
"""
strict_grammar = ('RINGInput', {
    'RINGInput': Either('Fragment', 'ReactionRule'),
    'Fragment': All('Prefix', Filler('fragment'), 'FragmentName',
                    Filler('{'), 'MolQuery', Filler('}')),
    'FragmentName': String(),
    'MolQuery': All('Atom', Optional('AtomChain')),
    'Prefix': All(Optional(Literals(['positive', 'negative', 'neutral'])),
                  Optional(Literals(['aromatic', 'olefinic', 'paraffinic'])),
                  Optional(Literals(['cyclic', 'linear']))),
    'Atom': All('AtomType', Filler('labeled'), 'AtomLabel',
                Optional(All(Filler('{'), 'AtomConstraintChain',
                             Filler('}')))),
    'AtomType': All(Optional('AtomPrefix'), 'Symbols', Optional('AtomSuffix')),
    'AtomPrefix': Literals(['aromatic', 'nonaromatic', 'ringatom',
                            'nonringatom', 'allylic']),
    'Symbols': Either(Literals(['any atom', '$', 'heteroatom',
                                '&', 'heavy atom', 'X']), String()),
    'AtomSuffix': Literals(['+', '-', '.', ':', '+.', '-.', '*']),
    'AtomLabel': String(),
    'AtomConstraintChain': All('AtomConstraints',
                               Optional(All(Filler(','),
                                            'AtomConstraintChain'))),
    'AtomConstraints': Either('AtomConstraintConnectivity',
                              'AtomConstraintRing'),
    'AtomConstraintConnectivity': All(Optional('Boolean'),
                                      Filler('connected to'),
                                      Optional('ConstraintNumber'),
                                      Either('GroupName', 'AtomType'),
                                      Optional(All(Filler('with'),
                                                   'BondType',
                                                   Filler('bond')))),
    'AtomConstraintRing': All(Optional('Boolean'),
                              Filler('in ring of size'),
                              'ConstraintNumber'),
    'Boolean': Literals(['!', '||', '&&', '+', '-']),
    'ConstraintNumber': All(Optional(Literals(['>', '=', '<', '>=', '<='])),
                            Digit()),
    'GroupName': All(Filler('group'), String()),
    'BondType': Literals(['single', 'double', 'triple', 'ring', 'nonring',
                          'aromatic', 'any', 'strong', 'partial']),
    'AtomChain': All(Either('BondedAtom', 'RingBond'),
                     Optional('AtomChain')),
    'BondedAtom': All('AtomType', Filler('labeled'), 'AtomLabel', 'BondType',
                      Filler('bond to'), 'AtomLabel',
                      Optional(All(Filler('{'), 'AtomConstraintChain',
                                   Filler('}')))),
    'RingBond': All(Filler('ringbond'), 'AtomLabel', 'BondType',
                    Filler('bond to'), 'AtomLabel'),
    'ReactionRule': All(Filler('rule'), 'ReactionName',
                        Filler('{'), 'Reactants',
                        Optional('Constraints'),
                        'TransformationChain', Filler('}')),
    'ReactionName': String(),
    'Reactants': All(Either('ReactantQuery', 'ReactantGroup', 'Duplicates'),
                     Optional('Reactants')),
    'ReactantQuery': All('Prefix', Filler('reactant'), 'ReactantName',
                         Filler('{'), 'MolQuery', Filler('}')),
    'ReactantName': String(),
    'ReactantGroup': All(Filler('reactant'), 'ReactantName', 'GroupName',
                         Filler('('), 'LabelMapping', Filler(')')),
    'Duplicates': All(Filler('reactant'), 'ReactantName',
                      Filler('duplicates'), 'ReactantName',
                      Filler('('), 'LabelMapping', Filler(')')),
    'LabelMapping': All('AtomLabel', Filler('=>'), 'AtomLabel',
                        Optional(All(Filler(','), 'LabelMapping'))),
    'Constraints': All(Filler('constraints{'), Optional('FragmentChain'),
                       'ConstraintChain', Filler('}')),
    'FragmentChain': All('Fragment', Optional('FragmentChain')),
    'ConstraintChain': All(Optional('Boolean'),
                           Either('BranchConstraint', 'Constraint'),
                           Optional(All('Boolean', 'ConstraintChain'))),
    'BranchConstraint': All(Filler('('),
                            Either('ConstraintChain', 'Constraint'),
                            Filler(')')),
    'Constraint': Either('C_Size', 'C_Charge', 'C_Cyclic', 'C_Characteristic',
                         'C_Fragment', 'C_Group'),
    'C_Size': All('SizeChain', 'ConstraintNumber'),
    'SizeChain': All('Size', Optional(All('Boolean', 'SizeChain'))),
    'Size': All('ReactantName', Filler('.size')),
    'C_Charge': All('ChargeChain', 'ConstraintNumber'),
    'ChargeChain': All('Charge', Optional(All('Boolean', 'ChargeChain'))),
    'Charge': All('ReactantName', Filler('.charge')),
    'C_Cylic': All('ReactantName', Filler('is cyclic')),
    'C_Characteristic': Either('C_Aromatic', 'C_Oxygenate', 'C_Heteroaromatic',
                               'C_Bridged', 'C_DeclaredCharacteristic',
                               'C_Smiles', 'C_Formula'),
    'C_Aromatic': All('ReactantName', Filler('is aromatic')),
    'C_Oxygenate': All('ReactantName', Filler('is oxygenate')),
    'C_Heteroaromatic': All('ReactantName', Filler('is heteroaromatic')),
    'C_Bridged': All('ReactantName', Filler('is bridged')),
    'C_Declaredcharacteristic': All('ReactantName', Filler('is'),
                                    'CharacteristicName'),
    'CharacteristicName': String(),
    'C_Smiles': All('ReactantName', Filler('is'), 'Smiles'),
    'Smiles': String(),
    'C_Formula': All('ReactantName', Filler('.formula is'),
                     'MolecularFormulaChain'),
    'MolecularFormulaChain': All('ElementSymbol', Optional(Number()),
                                 Optional('MolecularFormulaChain')),
    'C_Fragment': All('ReactantName', Filler('contains'),
                      Optional(All('ConstraintNumber', Filler('of'))),
                      'FragmentName'),
    'C_Group': All('ReactantName', Filler('contains'),
                   Optional(All('ConstraintNumber', Filler('of'))),
                   Optional(Filler('group')), 'GroupName'),
    'TransformationChain': All('ConnectivityChange',
                               Optional('TransformationChain')),
    'ConnectivityChange': Either('BondForm', 'BondBreak', 'BondModify',
                                 'BondDecrease', 'BondIncrease',
                                 'AtomTypeModify'),
    'BondForm': All(Filler('form'), Optional('BondType'), Filler('bond'),
                    Filler('('), 'AtomLabel', Filler(','), 'AtomLabel',
                    Filler(')')),
    'BondBreak': All(Filler('break'), Optional('BondType'), Filler('bond'),
                     Filler('('), 'AtomLabel', Filler(','), 'AtomLabel',
                     Filler(')')),
    'BondModify': All(Filler('modify bond'), Filler('('), 'AtomLabel',
                      Filler(','), 'AtomLabel', Filler(','),
                      'BondType', Filler(')')),
    'BondIncrease': All(Filler('increase bond order'), Filler('('),
                        'AtomLabel', Filler(','), 'AtomLabel',
                        Filler(')')),
    'BondDecrease': All(Filler('decrease bond order'), Filler('('),
                        'AtomLabel', Filler(','), 'AtomLabel',
                        Filler(')')),
    'AtomTypeModify': All(Filler('modify atomtype'), Filler('('),
                          'AtomLabel', Filler(','),
                          'AtomType', Filler(')'))
})

# Enhanced for radical electron related functions
enhanced_grammar = ('RINGInput', strict_grammar[1].copy())
enhanced_grammar[1]['BondType'] = Literals(['single', 'double', 'triple',
                                            'quadruple', 'ring', 'nonring',
                                            'aromatic', 'any', 'strong',
                                            'partial'])
enhanced_grammar[1]['AtomConstraints'] =\
    Either('AtomConstraintConnectivity', 'AtomConstraintRing',
           'AtomConstraintRadical', 'AtomConstraintNRing')
enhanced_grammar[1]['AtomConstraintRadical'] =\
    All(Optional('Boolean'), Filler('has'), 'ConstraintNumber',
        Filler('radical electrons'))
enhanced_grammar[1]['ConnectivityChange'] =\
    Either('BondForm', 'BondBreak', 'BondModify', 'BondDecrease',
           'BondIncrease', 'AtomTypeModify', 'RadicalModify',
           'RadicalIncrease', 'RadicalDecrease', 'ChargeIncrease',
           'ChargeDecrease')
enhanced_grammar[1]['RadicalModify'] =\
    All(Filler('modify number of radical'), Filler('('), 'AtomLabel',
        Filler(','), Number(), Filler(')'))
enhanced_grammar[1]['RadicalIncrease'] =\
    All(Filler('increase number of radical'), Filler('('),
        'AtomLabel', Filler(')'))
enhanced_grammar[1]['RadicalDecrease'] =\
    All(Filler('decrease number of radical'), Filler('('),
        'AtomLabel', Filler(')'))
enhanced_grammar[1]['ChargeIncrease'] =\
    All(Filler('increase formal charge'), Filler('('),
        'AtomLabel', Filler(')'))
enhanced_grammar[1]['ChargeDecrease'] =\
    All(Filler('decrease formal charge'),
        Filler('('), 'AtomLabel', Filler(')'))
# For ?, any type of charge, and radicals are accepted.
enhanced_grammar[1]['AtomSuffix'] = Literals(['+', '-', '.', ':', '+.',
                                              '-.', '*', '?', ':.'])
# Stereochemsitry
enhanced_grammar[1]['AtomChain'] =\
    All(Either('BondedAtom', 'RingBond', 'StereoDoubleBond'),
        Optional('AtomChain'))
enhanced_grammar[1]['StereoDoubleBond'] =\
    All(Filler('stereo double bond'), 'AtomLabel',
        Optional('Boolean'), 'DoubleBondStereoType',
        Filler('to'), 'AtomLabel', Filler('for double bond between'),
        'AtomLabel', Filler('and'), 'AtomLabel')
enhanced_grammar[1]['DoubleBondStereoType'] =\
    Literals(['cis', 'trans', 'notspecified'])
# Number of ring constraint
enhanced_grammar[1]['AtomConstraintNRing'] =\
    All(Optional('Boolean'), Filler('in'), 'ConstraintNumber', Filler('ring'))


def update_names(rules):
    for name in rules:
        rules[name].set_name(name)


update_names(strict_grammar[1])
update_names(enhanced_grammar[1])



# -*- coding: utf-8 -*-
from pgradd.RINGParser.Reader import Read
from rdkit import Chem
import unittest


class TestRINGParser(unittest.TestCase):

    def test_molquery(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1, 0), (0, 1), (1, 2), (2, 1))))

    def test_double_triple_bond(self):
        testmol = Chem.MolFromSmiles('C=C-C#C')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 double bond to c1
            C labeled c3 single bond to c2
            C labeled c4 triple bond to c3
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1, 2, 3),)))

    def test_aromatic_bond(self):
        testmol = Chem.MolFromSmiles('c1ccccc1')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 aromatic bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (0, 5), (1, 0), (1, 2), (2, 1),
                                     (2, 3), (3, 2), (3, 4), (4, 3), (4, 5),
                                     (5, 4), (5, 0))))

    def test_ring_bond1(self):
        testmol = Chem.MolFromSmiles('CCC1CCC1')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 ring bond to c1
            C labeled c3 ring bond to c2
            C labeled c4 ring bond to c3
            ringbond c4 ring bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
		                     sorted(((2, 3, 4, 5), (2, 5, 4, 3),
                                     (3, 2, 5, 4), (3, 4, 5, 2),
                                     (4, 3, 2, 5), (4, 5, 2, 3),
                                     (5, 2, 3, 4), (5, 4, 3, 2))))

    def test_ring_bond2(self):
        testmol = Chem.MolFromSmiles('CCC1CCC1')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 ring bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((2, 3), (2, 5), (3, 2), (3, 4), (4, 3),
                                     (4, 5), (5, 4), (5, 2))))

    def test_non_ring_bond(self):
        testmol = Chem.MolFromSmiles('CCC1CCC1')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 nonring bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1))))

    def test_any_bond1(self):
        testmol = Chem.MolFromSmiles('CC=CC#C')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 any bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1),
                                     (2, 3), (3, 2), (3, 4), (4, 3))))

    def test_any_bond2(self):
        testmol = Chem.MolFromSmiles('CC=CC#C')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 any bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1),
                                     (2, 3), (3, 2), (3, 4), (4, 3))))

    def test_strong_bond(self):
        testmol = Chem.MolFromSmiles('CC=CC#C')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 strong bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1, 2), (2, 1), (3, 4), (4, 3))))

    def test_other_bond1(self):
        testmol = Chem.MolFromSmiles('C[CH2-]')
        s = """
        positive fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertEqual(match_index, ())

    def test_other_bond2(self):
        testmol = Chem.MolFromSmiles('C[CH2+]')
        s = """
        negative fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertEqual(match_index, ())

    def test_other_bond3(self):
        testmol = Chem.MolFromSmiles('C=CC')
        s = """
        olefinic fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1, 2), (2, 1))))

    def test_other_bond4(self):
        testmol = Chem.MolFromSmiles('C=CC')
        s = """
        paraffinic fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertEqual(match_index, ())

    def test_other_bond5(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        paraffinic fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1))))

    def test_other_bond6(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        linear fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1))))

    def test_other_bond7(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        cyclic fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertEqual(match_index, ())

    def test_other_bond8(self):
        testmol = Chem.MolFromSmiles('C1CCC1C')
        s = """
        cyclic fragment a{
            C labeled c1
            C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (0, 3), (1, 0), (1, 2),
                                     (2, 1), (2, 3), (3, 2),
                                     (3, 4), (3, 0), (4, 3))))

    def test_symbol_atomsuffix(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        fragment a{
            $ labeled c1
            $ labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (0, 3), (0, 4), (0, 5), (1, 0),
                                     (1, 2), (1, 6), (1, 7), (2, 1), (2, 8),
                                     (2, 9), (2, 10), (3, 0), (4, 0),
                                     (5, 0), (6, 1), (7, 1), (8, 2), (9, 2),
                                     (10, 2))))

    def test_other_bond9(self):
        testmol = Chem.MolFromSmiles('CCO')
        s = """
        fragment a{
            X labeled c1
            X labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (1, 0), (1, 2), (2, 1))))

    def test_other_bond10(self):
        testmol = Chem.MolFromSmiles('CCS')
        s = """
        fragment a{
            X labeled c1
            & labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1, 2),)))

    def test_atom_constraint1(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 single bond to c1 {connected to =2 C}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1), (2, 1))))

    def test_atom_constraint2(self):
        testmol = Chem.MolFromSmiles('CCC')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 single bond to c1 {connected to =1 C}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1, 0), (1, 2))))

    def test_atom_constraint3(self):
        testmol = Chem.MolFromSmiles('CC=C')
        s = """
        fragment a{
            C labeled c1
            C labeled c2 single bond to c1 {connected to =1 C with double bond}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1),)))

    def test_atom_constraint4(self):
        testmol = Chem.MolFromSmiles('CC=C')
        s = """
        fragment a{
            C labeled c1 {connected to >1 C with any bond}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((1,),)))

    def test_atom_constraint5(self):
        testmol = Chem.MolFromSmiles('CC=C')
        s = """
        fragment a{
            C labeled c1 {!connected to >1 C with any bond}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, ), (2, ))))

    def test_atom_constraint6(self):
        testmol = Chem.MolFromSmiles('CC1CCC1')
        s = """
        fragment a{
            C labeled c1 {!in ring of size >0}
            C labeled c2 single bond to c1 {in ring of size >0}
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1),)))

    def test_atom_prefix1(self):
        testmol = Chem.MolFromSmiles('CC1CCC1')
        s = """
        fragment a{
            nonringatom C labeled c1
            ringatom C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1),)))

    def test_atom_prefix2(self):
        testmol = Chem.MolFromSmiles('Cc1ccccc1')
        s = """
        fragment a{
            nonaromatic C labeled c1
            aromatic C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1),)))

    def test_atom_prefix3(self):
        testmol = Chem.MolFromSmiles('CC=C')
        s = """
        fragment a{
            C labeled c1
            allylic C labeled c2 single bond to c1
        }
        """
        molquery = Read(s)
        match_index = molquery.GetQueryMatches(testmol)
        self.assertListEqual(sorted(match_index),
                             sorted(((0, 1),)))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from .. Error import UnitsError

from .utils import is_zero
from .parser import eval_expr


__all__ = [
    'eval_quantity', 'eval_qty', 'Quantity', 'ArrayQuantity',
    'FundamentalUnits']


def eval_quantity(expr):
    """
    Evaluate string expression of physical quantity or units.

    If `expr` is not a string type, then `expr` is returned unchanged.

    Parameters
    ----------
    expr : str
        Describe physical quantity

    Returns
    -------
    qty : :class:`Quantity`
        Representation of physical quantity
    """
    if isinstance(expr, str):
        return eval_expr(expr)
    else:
        return expr


eval_qty = eval_quantity  # Abbreviation for above (gets used a lot).


class GenericQuantity(object):
    @staticmethod
    def _unpack_qty(other):
        if isinstance(other, Quantity):
            value = other.value
            units = other.units
        elif isinstance(other, ArrayQuantity):
            value = other.view(np.ndarray)
            units = other._units
        elif isinstance(other, (list, tuple)):
            value = np.array(other)
            units = FundamentalUnits.null()
        else:
            value = other
            units = FundamentalUnits.null()
        return (value, units)

    @classmethod
    def _build(cls, value, units):
        if not units:
            return value

        if isinstance(value, (float, int)):
            return Quantity(value, units)
        elif isinstance(value, np.ndarray):
            new_value = value.view(ArrayQuantity)
            new_value._units = units
            return new_value
        else:
            raise AssertionError(
                'Unknown value type to GenericQuantity._build(): %r' % value)

    """
    def __new__(cls, value, units):
        q = cls._numeric_class.__new__(cls, value)
        q._units = units
        return q
    """

    def in_units(self, units):
        """
        Convert to specified `units` and return numerical value.

        Parameters
        ----------
        units : str or instance of :class:`Quantity`
            Describe the units to convert to before formatting

        Returns
        -------
        formatted : int or float
            The numerical value of the converted quantity

        Raises
        ------
        UnitsError
            If `units` are incompatible with the units of this quantity.
        """
        value = self/eval_qty(units)
        if isinstance(value, GenericQuantity):
            raise UnitsError("Units of '%s' are incompatible with '%s'"
                             % (self, units))
        return value

    def has_units(self, units):
        """
        Test for compatibility with given `units`.

        Parameters
        ----------
        units : str or instance of :class:`Quantity`
            Describe the units this quantity must be convertible to

        Returns
        -------
        is_compatible : bool
            `True` if and only if conversion to `units` is possible.
        """
        (self_value, self_units) = self._unpack_qty(self)
        if isinstance(units, FundamentalUnits):
            return self_units == units
        else:
            return self_units == self._unpack_qty(eval_qty(units))[1]

    def __eq__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if not is_zero(other) and (not other_units or
                                   not self.has_units(other_units)):
            return False
        return self_value == other_value

    def __ne__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if not is_zero(other) and (not other_units or
                                   not self.has_units(other_units)):
            return True
        return self_value != other_value

#    def __lt__(self, other):
#        (self_value, self_units) = self._unpack_qty(self)
#        (other_value, other_units) = self._unpack_qty(other)
#        if (not is_zero(other_value) and
#                (not other_units or not self.has_units(other_units))):
#            raise UnitsError(
#                'Incompatible units %s vs %s in comparison'
#                % (self_units, other_units))
#        return self_value < other_value

    def __lt__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in comparison'
                % (self_units, other_units))
        return self_value > other_value

    def __ge__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in comparison'
                % (self_units, other_units))
        return self_value >= other_value

    def __le__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in comparison'
                % (self_units, other_units))
        return self_value <= other_value

    def __add__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in addition'
                % (self_units, other_units))
        return self._build(self_value + other_value, self_units)

    def __radd__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in addition'
                % (self_units, other_units))
        return self._build(other_value + self_value, self_units)

    def __sub__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in subtraction'
                % (self_units, other_units))
        return self._build(self_value - other_value, self_units)

    def __rsub__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if (not is_zero(other_value) and
                (not other_units or not self.has_units(other_units))):
            raise UnitsError(
                'Incompatible units %s vs %s in subtraction'
                % (self_units, other_units))
        return self._build(other_value - self_value, self_units)

    def __mul__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        return self._build(self_value*other_value, self_units*other_units)

    def __rmul__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        return self._build(other_value*self_value, other_units*self_units)

    def __truediv__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        return self._build(self_value/other_value, self_units/other_units)
    __floordiv__ = __truediv__

    def __div__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        return self._build(self_value/other_value, self_units/other_units)
    __truediv__ = __div__

    def __rdiv__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        return self._build(other_value/self_value, other_units/self_units)
    __rtruediv__ = __rdiv__

    def __pow__(self, other):
        (self_value, self_units) = self._unpack_qty(self)
        (other_value, other_units) = self._unpack_qty(other)
        if other_units:
            raise TypeError('Invalid operation: exponentiation by quantity.')
        return self._build(
            self_value**other_value, self_units**other_value)

    def __rpow__(self, other):
        raise TypeError('Invalid operation: exponentiation by quantity.')

    def __neg__(self):
        (self_value, self_units) = self._unpack_qty(self)
        return self._build(-self_value, self_units)

    def __abs__(self):
        (self_value, self_units) = self._unpack_qty(self)
        return self._build(abs(self_value), self_units)

    def __str__(self):
        (value, units) = self._unpack_qty(self)
        return str(value) + ' ' + str(units)

    def __repr__(self):
        (value, units) = self._unpack_qty(self)
        return '%s(%r, %r)' % (type(self).__name__, value, units)


class Quantity(GenericQuantity):
    """
    Represent a physical quantity.

    .. note::
        This class should not be instantiated directly by users.  Rather, users
        should construct quantities and units using :func:`eval_qty()` or
        another helper function.

    Physical quantities usually possess units of some kind, and these are
    stored internally as an instance of :class:`FundamentalUnits`.
    Effectively, physical quantities are always stored in SI units to simplify
    arithmetic.  Various method are available to format or numerically convert
    them to any available compatible units.

    To facilitate meaningful computation with unitful quantities, this
    representation overloads several arithmetic and comparison operators:

        * Comparison operators (``==``, ``!=``, ``<``, ``<=``, ``>=``, and
          ``>``) are supported between quantities with compatible units.
        * Addition (``+``) and subtraction (``-``) are supported between
          quantities with compatible units.
        * Multiplication (``*``) and division (``/``) are supported between
          two quantities and between a scalar and a quantity.
        * Exponentation (``**``) is supported for a quantity as the base and a
          scalar as the exponent.
        * Conversion to type `bool`: The converted value is `True`, if and only
          if the quantity is non-zero.
        * Conversion to type `int` or type `float`: The converted value is
          the value of the quantity in equilvalent SI units.

    If two quantities involved have incompatible units and such compatiblity
    is required, the operation raises :exc:`chemtk.error.UnitsError`.

    If an arithmetic operation leads to a unitless result, then a number will
    be returned instead of a :class:`Quantity`.
    """
    __array_priority__ = 1.0

    def __init__(self, value, units):
        self.value = value
        if isinstance(units, str):
            units = eval_qty(units).units
        self.units = units

    def __hash__(self):
        return hash(self.value)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def fmt_in_units(self, units):
        """
        Format in specified `units`.

        Convert this quantity to `units` and return a string representation of
        the converted quantity.

        Parameters
        ----------
        units : str or instance of :class:`Quantity`
            Describe the units to convert to before formatting

        Returns
        -------
        formatted : str
            String representation of quantity after converting to `units`

        Raises
        ------
        UnitsError
            If `units` are incompatible with the units of this quantity.
        """
        return '%g %s' % (self.in_units(units), units)


class ArrayQuantity(GenericQuantity, np.ndarray):
    def __new__(cls, data, *args, **kwargs):
        units = kwargs.pop('units', None)
        if units:
            units = eval_qty(units).units
        if(hasattr(data, '__iter__')
                and all((isinstance(datum, GenericQuantity) or datum == 0)
                        for datum in data)):
            # Bundled together quantities.
            values = []
            use_units = None
            for datum in data:
                (value, datum_units) = cls._unpack_qty(datum)
                if not use_units and datum_units:
                    use_units = datum_units
                if(use_units and datum_units and use_units != datum_units
                        and value != 0 and not units):
                    raise UnitsError(
                        'Inconsistent units in contents provided to '
                        'ArrayQuantity initializer')
                values.append(value)

            data = values
            if units:
                if use_units != units:
                    raise UnitsError(
                        'Units in contents do not match specified units in '
                        'ArrayQuantity initializer')
            else:
                units = use_units

        arrqty = np.array(data, *args, **kwargs)
        if not issubclass(arrqty.dtype.type, np.number):
            raise TypeError('Units may only be assigned to numerical values.')

        arrqty = arrqty.view(cls)
        if units:
            arrqty._units = units
        elif hasattr(arrqty, '_units'):
            arrqty._units = arrqty._units
        else:
            raise TypeError(
                'Units must be specified when initializing ArrayQuantity.')
        return arrqty

    def __array_finalize__(self, arrqty):
        if hasattr(arrqty, '_units'):
            self._units = arrqty._units

    def __array_wrap__(self, outarr, context=None):
        if not issubclass(outarr.dtype.type, np.number):
            return np.asarray(outarr)
        return outarr

    def __getitem__(self, idx):
        result = np.ndarray.__getitem__(self, idx)
        if not isinstance(result, np.ndarray):
            return Quantity(result, self._units)
        return result

    def fmt_in_units(self, units):
        """
        Format in specified `units`.

        Convert this quantity to `units` and return a string representation of
        the converted quantity.

        Parameters
        ----------
        units : str or instance of :class:`Quantity`
            Describe the units to convert to before formatting

        Returns
        -------
        formatted : str
            String representation of quantity after converting to `units`

        Raises
        ------
        UnitsError
            If `units` are incompatible with the units of this quantity.
        """
        return '%s %s' % (self.in_units(units), units)


class FundamentalUnits(object):
    """
    Represent the fundamental units of a physical quantity.

    .. note::
        This class should not be instantiated directly by users.  Rather, users
        should construct quantities and units using :func:`eval_qty()` or
        another helper function.

    Units are represented internally as an array of numerical exponents, one
    each for the standard SI units:

        * meter (m)
        * kilogram (kg)
        * second (s)
        * ampere (A)
        * kevlin (K)
        * gram mole (mol)
        * candela (cd)

    The representation overloads multiplication (``*``), division (``/``),
    and equality (``==``) and inequality (``!=``) comparisons for use between
    instances of :class:`FundamentalUnits`.  Additionally, the exponentation
    (``**``) operator is suported between a :class:`FundamentalUnits` instance
    and a numerical exponent.  The truth value (*i.e.* ``bool(units)``) of an
    instance of :class:`FundamentalUnits` is `True` if and only if at least one
    exponent is non-zero.

    This abstraction allows for floating point numbers on the exponents,
    however, exponent values within a *threshold* (currently 10^-7) of an
    integer value are rounded to avoid comparison failures arising due to the
    inexactness of floating point arithmetic.
    """
    THRESHOLD_INTEGER = 1e-7
    _primitive_units = ['m', 'kg', 's', 'A', 'K', 'mol', 'cd']
    _primitive_units_index = dict((unit, i)
                                  for (i, unit) in enumerate(_primitive_units))

    @classmethod
    def new(cls, name):
        exps = np.zeros(len(cls._primitive_units))
        are_floats = np.zeros(len(cls._primitive_units), dtype='bool')
        exps[cls._primitive_units_index[name]] = 1
        return cls(exps, are_floats)

    @classmethod
    def null(cls):
        exps = np.zeros(len(cls._primitive_units))
        are_floats = np.zeros(len(cls._primitive_units), dtype='bool')
        return cls(exps, are_floats)

    @classmethod
    def _build(cls, exps):
        exps_rounded = exps.round()
        are_floats = np.abs(exps - exps_rounded) > cls.THRESHOLD_INTEGER
        return cls(np.select(
            [are_floats, np.logical_not(are_floats)],
            [exps, exps_rounded]), are_floats)

    def __init__(self, exps, are_floats):
        self.exps = exps
        self.are_floats = are_floats

    def __mul__(self, other):
        return self._build(self.exps + other.exps)

    def __truediv__(self, other):
        return self._build(self.exps - other.exps)

    def __div__(self, other):
        return self._build(self.exps - other.exps)

    def __pow__(self, other):
        return self._build(other*self.exps)

    def __eq__(self, other):
        print(self.exps, other.exps)
        return (self.exps == other.exps).all()

    def __ne__(self, other):
        return not (self == other)

    def __bool__(self):
        return bool(self.exps.any())

    def __nonzero__(self):
        return bool(self.exps.any())

    def __str__(self):
        up = []
        dn = []
        print('here')
        exps = self.exps
        for i, unit in enumerate(self._primitive_units):
            if exps[i] == 1:
                up.append(unit)
            elif exps[i] > 0:
                if self.are_floats[i]:
                    up.append('%s^%s' % (unit, exps[i]))
                else:
                    up.append('%s^%s' % (unit, int(exps[i])))
            elif exps[i] == -1:
                dn.append(unit)
            elif exps[i] < 0:
                if self.are_floats[i]:
                    dn.append('%s^(%s)' % (unit, exps[i]))
                else:
                    dn.append('%s^(%s)' % (unit, int(exps[i])))
        if not up:
            s = '1'
        else:
            s = '*'.join(up)
        if not dn:
            return s
        elif len(dn) == 1:
            return s + '/' + dn[0]
        else:
            return s + '/(' + '*'.join(dn) + ')'

    def __repr__(self):
        return "<%s: '%s'>" % (type(self).__name__, str(self))



import sys


__all__ = [
    'string', 'SchemaDefinitionError', 'InputDataError', 'InputDataWarning']


if sys.version_info[0] >= 3:
    string = str
else:
    string = str
    # string = unicode


class SchemaDefinitionError(Exception):
    pass


class InputDataError(Exception):
    pass


class InputDataWarning(Warning):
    pass



from collections.abc import Mapping, Sequence
from warnings import warn

from .common import string, SchemaDefinitionError, InputDataError,\
    InputDataWarning
from .lib_interface import YAMLTaggedValue


__all__ = ['SchemaRepository', 'SchemaDefinitionError']


class AnonymousClass(Mapping):
    def __contains__(self, key):
        return key in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self):
        return iter(self.__dict__)

    def __getitem__(self, item):
        return getattr(self, item)


def _get_type_and_attrs(schema_def):
    if schema_def is None:
        return ('any', {})
    elif isinstance(schema_def, string):
        return (schema_def, {})
    elif isinstance(schema_def, Mapping):
        attrs = schema_def.copy()
        if 'type' not in attrs:
            raise SchemaDefinitionError("missing 'type' attribute in "
                                        "schema definition %r" % schema_def)
        del attrs['type']
        return (schema_def['type'], attrs)
    else:
        raise SchemaDefinitionError("invalid schema definition %r"
                                    % schema_def)


class ListLoader(object):
    def __init__(self, repo, item_type, object_class=None, tag=None):
        self.object_class = object_class
        self.tag = tag
        self.item_loader = repo.make_loader(item_type)

    def __call__(self, name, data, context):
        # params = {}
        if isinstance(data, YAMLTaggedValue):
            # Check proper subclass ...
            tag = data.tag
            data = data.value
            if self.tag is not None and tag != self.tag:
                repo = context['repo']
                if repo.is_tag_sub_type(tag, self.tag):
                    return repo.load_tagged(name, data, context, tag)
                else:
                    raise InputDataError(
                        "Expected %r object (or sub type), "
                        "got %r object" % (self.tag, tag))
        elif not isinstance(data, Sequence):
            raise InputDataError('Input data is not sequence type: %r' % data)

        # Load parameters with multiple alternatives.
        result = [self.item_loader(name, item, context) for item in data]

        # Return constructed value.
        if self.object_class is not None:
            # Class of object is defined.
            if hasattr(self.object_class, 'yaml_construct'):
                # User provided a constructor.
                try:
                    return self.object_class.yaml_construct(result, context)
                except Exception as exc:
                    from sys import exc_traceback
                    from traceback import format_tb
                    raise InputDataError(
                        '%s.yaml_construct() reported error '
                        '"%s" with traceback:\n%s'
                        % (self.object_class.__name__, exc,
                            ''.join(format_tb(exc_traceback))))
            else:
                return result
        else:
            return result


class ObjectLoader(object):
    def __init__(
            self, repo, members, object_class=None, tag=None,
            has_open_namespace=False):
        if not isinstance(members, Mapping):
            raise InputDataError('Provided members %r is not a mapping type'
                                 % members)

        self.object_class = object_class
        self.tag = tag
        self.has_open_namespace = has_open_namespace
        members = members.copy()

        # Convert parameter names to unicode strings.
        renames = []
        for name in members:
            if not isinstance(name, str):
                old_name = name
                renames.append((old_name, string(name)))
        for (old_name, new_name) in renames:
            entry = members[old_name]
            del members[old_name]
            members[new_name] = entry

        # Validate schema attributes and collect conflicts and alternatives.
        # conflicts = {}  -- TODO
        self.loaders = {}
        self.descs = {}
        self.requireds = set()
        self.optionals = set()
        self.defaults = {}
        self.alt_sets = set()
        alt_sets_index = {}

        for name in members:
            # Determine attributes of parameter.
            schema_def = members[name]
            typ, attrs = _get_type_and_attrs(schema_def)

            # Extract description of parameter (if one is available).
            self.descs[name] = attrs.get('desc')

            # Look for special attributes: optional, default, or alts:
            if attrs.get('optional'):
                if 'default' in attrs or attrs.get('alts'):
                    raise SchemaDefinitionError("use of 'default' or 'alts'",
                                                "attributes conflict with use",
                                                "of 'optional' in %r"
                                                % schema_def)
                self.optionals.add(name)
            elif 'default' in attrs:
                if attrs.get('optional') or attrs.get('alts'):
                    raise SchemaDefinitionError("use of 'optional' or 'alts'",
                                                "attributes conflicts with",
                                                "use of 'optional' in %r"
                                                % schema_def)
                self.defaults[name] = attrs['default']
            elif 'alts' in attrs:
                if attrs.get('optional') or 'default' in attrs:
                    raise SchemaDefinitionError("use of 'optional' or",
                                                "'default' attributes",
                                                "conflicts with use of 'alts'",
                                                "in %r" % schema_def)

                # Get list of alternative parameter names:
                alts = attrs['alts']
                if isinstance(alts, str):
                    alts = [alts]
                if not isinstance(alts, Sequence):
                    raise SchemaDefinitionError("'alts' attribute must be a",
                                                "sequence instead of %r."
                                                % alts)

                # Compute the complete set of alternatives.
                class hashable_set(set):
                    def __hash__(self):
                        return id(self)

                    def __eq__(self, other):
                        return id(self) == id(other)

                alt_set = hashable_set(alts)
                self.alt_sets.add(alt_set)
                alt_set.add(name)
                for alt in alts:
                    if alt in alt_sets_index:
                        if alt_sets_index[alt] in self.alt_sets:
                            self.alt_sets.remove(alt_sets_index[alt])
                        alt_set.update(alt_sets_index[alt])  # A = A U B

                # Make sure all alt_sets_index entries refer to the newly
                # created set.
                for alt in alt_set:
                    alt_sets_index[alt] = alt_set
            else:
                self.requireds.add(name)

            # Create new schema.
            self.loaders[name] = repo.make_loader(schema_def)

    def __call__(self, name, data, context):
        params = {}
        if isinstance(data, YAMLTaggedValue):
            # Check proper subclass ...
            tag = data.tag
            data = data.value
            if self.tag is not None and tag != self.tag:
                repo = context['repo']
                if repo.is_tag_sub_type(tag, self.tag):
                    return repo.load_tagged(name, data, context, tag)
                else:
                    raise InputDataError(
                        "Expected %r object (or sub type), "
                        "got %r object" % (self.tag, tag))
        # elif data is None:
        #    return None
        elif not isinstance(data, Mapping):
            raise InputDataError('Input data for %r is not mapping type: %r'
                                 % (name, data))

        # Check for parameters not defined in schema.
        for name in data:
            if name not in self.loaders and not self.has_open_namespace:
                warn('Invalid parameter %r specified in YAML input'
                     % name, InputDataWarning, 3)
                # raise InputDataError('Invalid parameter %r specified in %r'
                #    %(name, data))

        # Load required parameters.
        for name in self.requireds:
            if name not in data:
                raise InputDataError('Missing required parameter %r in %r'
                                     % (name, data))
            params[name] = self.loaders[name](name, data[name], context)

        # Load optional parameters.
        for name in self.optionals:
            if name in data:
                params[name] = self.loaders[name](name, data[name], context)

        # Load parameters with defaults.
        for name in self.defaults:
            if name in data:
                params[name] = self.loaders[name](name, data[name], context)
            else:
                params[name] = self.loaders[name](
                    name, self.defaults[name], context)

        # Load parameters with multiple alternatives.
        for alt_set in self.alt_sets:
            alt_names = [name for name in alt_set if name in data]
            if len(alt_names) != 1:
                raise InputDataError("Number of alternatives %r specified in",
                                     "%r does not equal 1" % (alt_set, data))
            name = alt_names[0]
            params[name] = self.loaders[name](name, data[name], context)

        # Return constructed value.
        if self.object_class is not None:
            # Class of object is defined.
            if hasattr(self.object_class, 'yaml_construct'):
                # User provided a constructor.
                params['name'] = name
                try:
                    return self.object_class.yaml_construct(params, context)
                except Exception as exc:
                    from sys import exc_traceback
                    from traceback import format_tb
                    raise InputDataError(
                        '%s.yaml_construct() reported error',
                        '"%s" with traceback:\n%s'
                        % (self.object_class.__name__, exc,
                            ''.join(format_tb(exc_traceback))))
            else:
                # Default constructor.
                obj = type(self.object_class.__name__, (object,), {})()
                for name in params:
                    setattr(obj, name, params[name])
                obj.__class__ = self.object_class
                return obj
        else:
            # Class of object is undefined.
            obj = AnonymousClass()
            for name in params:
                setattr(obj, name, params[name])
            return obj


class SchemaRepository(object):
    def __init__(self, builtins=True):
        if builtins:
            # Setup builtin schemata.
            from .builtins import loaders
            self._loaders = loaders
        else:
            # No builtin schemata.
            self._loaders = {}
        self._loaders['object'] = ObjectLoader

    def make_loader(self, schema_def):
        """Return a function that accepts an abstract tree conforming to the
        provided schema definition and constructs a Python object.
        """
        type_name, attrs = _get_type_and_attrs(schema_def)
        loader_class = self._loaders.get(type_name)
        if loader_class is None:
            raise SchemaDefinitionError('undefined data type %r' % type_name)

        def del_key(attrs, key):
            if key in attrs:
                del attrs[key]
        for key in ['optional', 'default', 'alts', 'type', 'desc']:
            del_key(attrs, key)
        try:
            return loader_class(self, **attrs)
        except TypeError:
            import sys
            from traceback import extract_tb
            typ, value, tb = sys.exc_info()
            mod_name, lineno, func_name, code = extract_tb(tb)[-1]
            if(mod_name == __file__
                    and func_name == sys._getframe().f_code.co_name):
                raise SchemaDefinitionError(
                    'Invalid attributes to type %r: %r' % (type_name, attrs))
            else:
                raise

    class ObjectLoaderFactory(object):
        def __init__(self, members, object_class=None, tag=None):
            self.members = members
            self.object_class = object_class
            self.tag = tag

        def __call__(self, repo, **attrs):
            return ObjectLoader(repo, self.members, self.object_class,
                                self.tag, **attrs)

    class ListLoaderFactory(object):
        def __init__(self, item_type, object_class=None, tag=None):
            self.item_type = item_type
            self.object_class = object_class
            self.tag = tag

        def __call__(self, repo, **attrs):
            return ListLoader(repo, self.item_type, self.object_class,
                              self.tag, **attrs)

    def load_tagged(self, name, data, context, tag):
        loader = self._loaders.get(tag[1:])
        if loader is None:
            raise InputDataError('Unknown tag: %r' % tag)
        return loader(self)(name, data, context)

    def is_tag_sub_type(self, tag, base_tag):
        loader = self._loaders.get(tag[1:])
        base_loader = self._loaders.get(base_tag[1:])
        if loader is None:
            raise InputDataError('Invalid tag: %r' % tag)
        if base_loader is None:
            raise InputDataError('Invalid tag: %r' % base_tag)
        object_class = loader.object_class
        base_object_class = base_loader.object_class
        if object_class is not None and base_object_class is not None:
            return issubclass(object_class, base_object_class)

    def make_object_loader(self, members, object_class=None,
                           open_namespace=False):
        """Return a function that accepts an abstract tree conforming to a
        schema for an object whose `members` are defined by a yaml mapping.
        """
        return self.ObjectLoaderFactory(members, object_class)(
            self, has_open_namespace=open_namespace)

    def register_class(self, name, schema_def, class_):
        """Register a Python class to be used to construct objects having
        specified YAML tag (`name`) with provided schema definition.
        """
        self.register_type(name,
                           self.ObjectLoaderFactory(schema_def, class_,
                                                    tag='!' + name))

    def register_list_type(self, name, item_type, class_):
        """Register a Python class to be used to construct objects having
        specified YAML tag (`name`) in which the data is expected to be a list
        containing items of specified `item_type`.
        """
        self.register_type(name,
                           self.ListLoaderFactory(item_type,
                                                  class_, tag='!' + name))

    def register_type(self, name, loader):
        """Register a function to be called (`loader`) to construct new
        objects from YAML data tagged with `name`.
        """
        self._loaders[name] = loader

    def unregister_type(self, name):
        del self._schemata[name]



"""
=======================================
Physical quantities and unit conversion
=======================================

This module provides routines for representing and manipulating physical
quantities with units.  A physical quantity is represented internally by its
numerical value in SI units combined with a :class:`FundamentalUnits` that
specifies the exponents of meters, kilograms, seconds, ampere, kelvin,
gram-moles, and candela represented in the quantity.

All standard prefixes (*e.g.* k for kilo, u for micro, etc.) are supported as
are most derived units in the metric and standard systems.  Notable absences
include celsius and fahrenheit as these do not represent temperatures in the
absolute sense.  Proper handling of these units in arithmetic is complex and
beyond the intended scope of this library.


-------
Summary
-------

.. autosummary::

    eval_quantity
    eval_qty
    Quantity
    FundamentalUnits
    with_units
    in_units
    has_units
    from_SI_to
    to_SI_from


--------
Examples
--------

>>> from chemtk.units import with_units, in_units, has_units
>>> g = with_units(9.81, 'm/s^2')
>>> m = with_units(10.0, 'lb')
>>> print m*g
44.497411497 m*kg/s^(-2)
>>> print in_units(m, 'g')
4535.9237
>>> print (m*g).fmt_in_units('N')
44.4974 N
>>> print (m*g).fmt_in_units('lbf')
10.0034 lbf
>>> print m.has_units('ug')
True
>>> print m.has_units('J s')
False
>>>


---------
Reference
---------

.. autodoc puts stuff extracted from docstrings here.
"""

from .qty import *
from .helpers import *
from . import builtin

__all__ = qty.__all__[:] + helpers.__all__[:]

del qty, helpers, builtin



# 2016. Vlachos Group Geun Ho Gu. University of Delaware.
"""
=============================
Group-additivity computations
=============================

This module is used for using Group additivity method to compute energies

--------
Examples
--------

>>> from pgradd.GroupAdd.Library import GroupLibrary
>>> import pgradd.ThermoChem
>>> lib = GroupLibrary.Load('benson')
>>> groups = lib.GetDescriptors('CC')
>>> print groups
>>> thermochem = lib.Estimate(groups,'thermochem')
>>> print thermochem.get_HoRT(298.15)
defaultdict(<type 'int'>, {'C(C)(H)3': 2})
-34.4280812417

"""

from . Library import GroupLibrary
from . Scheme import GroupAdditivityScheme

__all__ = ['GroupLibrary', 'GroupAdditivityScheme']



# 2016. Vlachos Group Geun Ho Gu. University of Delaware.


####
#
# setuptools likes to see a name for the package,
# and it's best-practices to have the __version__
# present, too:
#
name = 'pgradd'
__version__ = '2.9.6'
#
####



import re
from collections import defaultdict

from .. Error import GroupSyntaxError


# Note: This class is not currently documented explicitly and is only used to
# inherit Group() from.
class Descriptor(object):
    """
    Represent a chemical descriptor upon which properties may
    linearly depend on.
    """
    def __init__(self, scheme, name):
        """Initialize a chemical descriptor with given string name.

        Parameters
        ----------
        scheme : instance of :class:`GroupScheme`
            Specify group-additivity scheme for which the group belongs to.
        name : str
            Specify name of descriptor
        """
        self.scheme = scheme
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.name == other.name
        else:
            return self.name == other

    def __neq__(self, other):
        return not (self == other)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s: %s>' % (type(self).__name__, str(self))


class Group(Descriptor):
    """Represent a chemical functional group.

    This object is hashable (meaning that it may be used as a dictionary key)
    and can be compared for equality against others of its kind, regardless of
    the order in which each the peripheral subgroups were specified at
    initialization.

    In addition to the default constructor (:meth:`Group.__init__()`), one can
    instantiate :class:`Group` objects using the alternative constructor
    :meth:`Group.parse()` which accepts a string name for the group that is
    parsed to determine the contained `csg` and `psgs`.
    """
    _parser_re = re.compile('[()]')

    @classmethod
    def parse(cls, scheme, text):
        """Initialize a chemical functional group given a string name.

        The syntax consists of the center-subgroup name followed by one or more
        peripheral subgroup names enclosed in parantheses, each of which is
        optionally followed by a numerical repeat count.

        Parameters
        ----------
        text : str
            Name of the group

        Examples
        --------
        Using the (builtin) ``'benson'`` scheme::

            >>> from chemtk.groupadd import Group, GroupScheme
            >>> scheme = GroupScheme.load('benson')
            >>> Group.parse(scheme, 'C(C)(H)3')
            Group('C', ['C', 'H', 'H', 'H'])
            >>> Group.parse(scheme, 'C(C[d])(C)2(H)')
            Group('C', ['C[d]', 'C', 'C', 'H'])
            >>> Group.parse(scheme, 'O(C)(H)')
            Group('O', ['C', 'H'])
        """
        def append_to_psgs(count, psg):
            for i in range(count):
                psgs.append(psg)
        parts = cls._parser_re.split(text)
        csg = parts[0]
        next_psg = None  # next nearest neighbor to be appended to nns
        psgs = []  # list of nearest neighbors
        for part in parts[1:]:
            if not part:
                continue
            if part.isdigit():
                if next_psg is None:
                    raise GroupSyntaxError('Expected base atom for Benson',
                                           'group got number instead.')
                append_to_psgs(int(part), next_psg)
                next_psg = None
            else:
                if next_psg is not None:
                    append_to_psgs(1, next_psg)
                    next_psg = None
                next_psg = part
        if next_psg is not None:
            append_to_psgs(1, next_psg)
        return cls(scheme, csg, psgs)

    def __init__(self, scheme, csg, psgs):
        """Initialize a chemical functional group.

        Parameters
        ----------
        scheme : instance of :class:`GroupScheme`
            Specify group-additivity scheme for which the group belongs to.
        csg : str
            Name of center-subgroup
        psgs : iterable of strs
            Names of peripheral-subgroups
        """
        self.csg = csg
        self.psgs = psgs
        canon_name = self._canonical_name()
        Descriptor.__init__(self, scheme, canon_name)

    def _canonical_name(self):
        """Return canonical name of group.

        Parameters
        ----------
        scheme : :class:`GroupScheme`
            Specify group-additivity scheme for which the group belongs to.

        Returns
        -------
        name : str
            Canonical name of group
        """
        canon_name = self.csg
        psg_counts = defaultdict(int)
        for psg in self.psgs:
            psg_counts[psg] += 1
        for name in sorted(psg_counts):
            if psg_counts[name] == 1:
                canon_name += '(' + name + ')'
            else:
                canon_name += '(' + name + ')' + '%d' % psg_counts[name]
        return canon_name

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.csg, self.psgs)



from .utils import is_zero
from .qty import eval_qty


__all__ = ['with_units', 'in_units', 'has_units', 'to_SI_from', 'from_SI_to']


def with_units(number, units):
    """
    Construct a physical quantity from the given `number` and `units`.

    If `number` is `None`, then return `None`, regardless of `units`.

    .. note::
        If `number` is not `None`, then the following two expressions are
        equivalent:

            1. ``chemtk.units.with_units(number, units)``
            2. ``number*chemtk.units.eval_qty(units)``

    Parameters
    ----------
    number : int, float, or NoneType
        Specify the numerical part of the new quantity
    units : str or :class:`Quantity`
        Describe the units of the new quantity

    Returns
    -------
    qty : :class:`Quantity` or NoneType
        The new quantity if `number` is not None; otherwise `None`
    """
    if number is None:
        return None
    if is_zero(number):
        return number
    return number*eval_qty(units)


def in_units(qty, units):
    """
    Convert quantity to specified `units` and return numerical value.

    If `qty` is `None`, then return `None`, regardless of `units`.

    .. note::
        If `qty` is not `None`, then the following two expressions are
        equivalent:

            1. ``chemtk.units.in_units(qty, units)``
            2. ``qty.in_units(units)``

    Parameters
    ----------
    qty : :class:`Quantity` or NoneType
        Specify the quantity to convert
    units : str or :class:`Quantity`
        Describe the units to convert `qty` to

    Returns
    -------
    number : int, float, or NoneType
        The numerical value of the converted quantity, if `qty` is not None;
        otherwise `None`
    """
    if qty is None:
        return None
    return qty.in_units(units)


def has_units(qty, units):
    """
    Test quantity (`qty`) for compatibility with given `units`.

    .. note::
        The following two expressions are equivalent:

            1. ``chemtk.units.has_units(qty, units)``
            2. ``qty.has_units(units)``

    Parameters
    ----------
    qty : :class:`Quantity` or NoneType
        Specify the quantity to test
    units : str or :class:`Quantity`
        Describe the units `qty` must be convertible to

    Returns
    -------
    is_compatible : bool
        True if and only if the units of `qty` can be converted to `units`.
    """
    return qty.has_units(units)


def to_SI_from(value, units):
    """
    Convert `value` having specified `units` to the equivalent SI units.

    Parameters
    ----------
    value : int or float
        Specify value of quantity to convert to SI units
    units : str or :class:`Quantity`
        Describe the units associated with `value`

    Returns
    -------
    new_value : int or float
        Numerical value of SI units quantity
    """
    return value*eval_qty(units).value


def from_SI_to(value, units):
    """
    Convert `value` in SI units to specified compatible `units`.

    Parameters
    ----------
    value : int or float
        Specify value of SI units quantity to convert
    units : str or :class:`Quantity`
        Describe the units to convert SI quantity to

    Returns
    -------
    new_value : int or float
        Numerical value of converted quantity.
    """
    return value/eval_qty(units).value



from ..Error import RINGSyntaxError
from operator import eq

"""
Parses the RING input string into abstract syntax tree as defined by
Grammar.py.

Example:
>>>from pgradd.RINGParser.parser import parse
>>>parse('\
>>>fragment C3Chain{\
>>>   C labeled c1\
>>>   C labeled c2 single bond to c1\
>>>   C labeled c3 single bond to c2\
>>>   }')
Out:
[RINGToken('RINGInput'),
 [RINGToken('Fragment'),
  [RINGToken('FragmentName'), 'C3Chain'],
  [RINGToken('MolQuery'),
   [RINGToken('Atom'),
    [RINGToken('AtomType'), [RINGToken('Symbols'), 'C']],
    [RINGToken('AtomLabel'), 'c1']],
   [RINGToken('AtomChain'),
    [RINGToken('BondedAtom'),
     [RINGToken('AtomType'), [RINGToken('Symbols'), 'C']],
     [RINGToken('AtomLabel'), 'c2'],
     [RINGToken('BondType'), 'single'],
     [RINGToken('AtomLabel'), 'c1']],
    [RINGToken('AtomChain'),
     [RINGToken('BondedAtom'),
      [RINGToken('AtomType'), [RINGToken('Symbols'), 'C']],
      [RINGToken('AtomLabel'), 'c3'],
      [RINGToken('BondType'), 'single'],
      [RINGToken('AtomLabel'), 'c2']]]]]]]
"""


class RINGToken(object):
    # Each node is converted to RINGToken
    def __init__(self, name):
        self.name = name

# TODO: Resolve unknown name 'cmp'

    def __cmp__(self, other):
        if isinstance(other, RINGToken):
            return eq(self.name, other.name)
        else:
            return eq(self.name, other)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "RINGToken('%s')" % self.name


class Parser(object):
    # Parser is a parent class for various type of syntax
    # syntax classes:
    # EOS               : end of the line identifier
    # Digit             : 1 digit class
    # Number            : mutiple digit class
    # String            : string class. (when to stop interpret is defined
    #                     by string_stop
    # Literal           : Has to match the specified symbol to be identified
    #                     as a class
    # Filler            : Looks for the match, but the information is
    #                     not stored.
    # DeprecatedLiteral : similar to Literal, but warning is raised.
    # Optional          : Indiciates that child class of this class is
    #                     not necessary
    # All               : Interpreter goes through all the child classes
    #                     of this class
    # Either            : At least one of the child class has to be in
    #                     the stream
    # ZeroOrMore        : Similar to Optional but can take multiple input
    # Literals          : Has to match one of the strings in this class.
    #                     Equal to Either(Literal('something'),
    #                     Literal('Something'),...)
    #
    # Notes:
    # The code compares the stream with given syntax tree in the grammar. so
    # ParseState.parse() stacks with the number of child class. When the errors
    # occur, the parser undo each parse loop by using 'try'. Eventually,
    # The error is given at the user's calling function.
    # For classes like Either and Optional, the algorithm still uses errors
    # when there is no match, but the error is suppressed using 'with'
    # function, A lot of valuable coding is in this work.
    def set_name(self, name):
        self.name = name


class EOS(Parser):
    def __call__(self, stream, output):
        if stream.peek() != '':
            stream.error('<end of string>')

    def __str__(self):
        return '<end of string>'

    def __repr__(self):
        return 'EOS()'


class Digit(Parser):
    def __init__(self, n=1):
        self.n = n

    def __call__(self, stream, output):
        out = ''
        for i in range(self.n):
            c = stream.peek()
            if not c.isdigit():
                stream.error('<digit>')
            out += stream.take()
        output.append(int(out))

    def __str__(self):
        if self.n == 1:
            return '<digit>'
        else:
            return '<%d digits>' % self.n

    def __repr__(self):
        if self.n == 1:
            return 'Digit()'
        else:
            return 'Digit(%r)' % self.n


class Number(Parser):
    def __init__(self):
        pass

    def __call__(self, stream, output):
        out = stream.peek()
        if not out.isdigit():
            stream.error('<number>')
        stream.take()
        while stream.peek().isdigit():
            out += stream.take()
        output.append(int(out))

    def __str__(self):
        return '<number>'

    def __repr__(self):
        return 'Number()'


string_okay = ['_']


class String(Parser):
    # Read continuous input of alphabet, digit, and any other character in
    # string_okay
    def __init__(self):
        pass

    def __call__(self, stream, output):
        if not (stream.peek().isalpha() or
                stream.peek().isdigit() or
                stream.peek() in string_okay):
                stream.error('<string>')
        nn = 2

        while (stream.peek(nn)[-1].isalpha() or
               stream.peek(nn)[-1].isdigit() or
               stream.peek(nn)[-1] in string_okay):
            nn += 1
        out = stream.take(n=nn-1)
        output.append(out)

    def __str__(self):
        return '<string>'

    def __repr__(self):
        return 'String()'


class Literal(Parser):
    def __init__(self, tok, no_error=False):
        self.tok = tok
        self.no_error = no_error

    def __call__(self, stream, output):
        if stream.peek(len(self.tok)) != self.tok:
            if self.no_error:
                # Suppress error message.  (It's handled by something further
                # up the stack).
                stream.error(None)
            else:
                stream.error("%r" % self.tok)
        else:
            stream.take(len(self.tok))
            output.append(self.tok)

    def __str__(self):
        return "%r" % self.tok

    def __repr__(self):
        return "Literal(%r)" % self.tok


class Filler(Parser):
    def __init__(self, tok, no_error=False):
        self.tok = tok
        self.no_error = no_error

    def __call__(self, stream, output):
        if stream.peek(len(self.tok)) != self.tok:
            if self.no_error:
                # Suppress error message.
                # (It's handled by RINGError in error.py)
                stream.error(None)
            else:
                stream.error("%r" % self.tok)
        else:
            stream.take(len(self.tok))

    def __str__(self):
        return "%r" % self.tok

    def __repr__(self):
        return "Literal(%r)" % self.tok


class DeprecatedLiteral(Literal):
    def __call__(self, stream, output):
        import warnings
        Literal.__call__(self, stream, output)
        warnings.warn("Use of '%s' is deprecated in RING inputs" % self.tok)

    def __str__(self):
        return "'%s' (deprecated)" % self.tok

    def __repr__(self):
        return "DeprecatedLiteral(%r)" % self.tok


class Optional(Parser):
    def __init__(self, opt):
        self.opt = opt

    def __call__(self, stream, output):
        with stream:
            stream.parse(self.opt, output)

    def __str__(self):
        return "%s (optional)" % self.opt

    def __repr__(self):
        return "Optional(%r)" % self.opt


class All(Parser):
    def __init__(self, *reqs):
        self.reqs = reqs

    def __call__(self, stream, output):
        for req in self.reqs:
            stream.parse(req, output)

    def __str__(self):
        return '<%s>' % ' '.join('%s' % req for req in self.reqs)

    def __repr__(self):
        return 'All(%s)' % ', '.join('%r' % req for req in self.reqs)


class Either(Parser):
    def __init__(self, *alts):
        self.alts = alts

    def __call__(self, stream, output):
        for alt in self.alts:
            with stream:
                stream.parse(alt, output)
                return
        raise stream.current_error

    def __str__(self):
        return '<%s>' % ' | '.join('%s' % (alt,) for alt in self.alts)

    def __repr__(self):
        return 'Either(%s)' % ', '.join("%r" % (alt,) for alt in self.alts)


class ZeroOrMore(Parser):
    def __init__(self, what):
        self.what = what

    def __call__(self, stream, output):
        with stream:
            stream.parse(self.what, output)
        while not stream.has_error:
            with stream:
                stream.parse(self.what, output)

    def __str__(self):
        return '%s (zero or more)' % self.what

    def __repr__(self):
        return 'ZeroOrMore(%r)' % self.what


class Literals(Either):
    def __init__(self, literal_strings):
        # Sort literals so that parser attempts to match longer literals first.
        # i.e.:  'Cl' should match chlorine and not carbon (with an 'l' left
        # over)
        sorted_strings = sorted(
            literal_strings, key=lambda s: len(s), reverse=True)
        Either.__init__(self, *(Literal(s, no_error=True)
                        for s in sorted_strings))

    def __call__(self, stream, output):
        try:
            return Either.__call__(self, stream, output)
        except RINGSyntaxError:
            if hasattr(self, 'name'):
                stream.error('<' + self.name + '>')
            else:
                raise

# filler is used in parserstate.take() which will move pointers
# beyond the following symbols.


filler = [' ', '\n', '\t']


class ParseState(object):
    """Parse provided stream into an Abstract Syntax Tree."""
    def __init__(self, grammar, stream, debug=False):
        """Parse provided stream into an Abstract Syntax Tree."""
        self.root, self.rules = grammar
        self.stream = stream
        self.debug = debug

        self.stack = []
        self.has_error = False
        self.current_error = None
        self.sidx = 0    # position in the actual string.
        self.lineno = 1  # Human readable pointer for line number
        self.colno = 1   # Human readable pointer for column number
        self.skip_filler()
        if self.debug:
            self.depth = 0

    def error(self, mesg):
        """Raise a RINGSyntaxError exception with the given message.  Line
        and column numbers are provided to the exception object automatically.
        """
        raise RINGSyntaxError(mesg, self.lineno, self.colno, self.stream)

    def peek(self, n=1):
        """Return the next n characters in the character stream.  Do not
        update the stream pointer.
        """
        return self.stream[self.sidx:self.sidx + n]

    def take(self, n=1):
        """Return the next n characters in the character stream and update the
        stream pointer.
        """
        # Move pointer to after the space.
        s = self.stream[self.sidx:self.sidx + n]
        for chr in s:
            if chr == '\n':  # if string contains enter, move to next line.
                self.lineno += 1
                self.colno = 1
            else:
                self.colno += 1
        self.sidx += n
        # skip '\n' and  ' '
        self.skip_filler()

        return s

    def skip_filler(self):
        while self.peek() in filler:
            if self.peek() == '\n':
                self.lineno += 1
                self.colno = 1
                self.sidx += 1
            else:
                self.colno += 1
                self.sidx += 1

    def __enter__(self):
        self.stack.append((self.sidx, self.lineno, self.colno))
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            (self.sidx, self.lineno, self.colno) = self.stack.pop()

            if exc_type is RINGSyntaxError:
                if self.current_error is not None:
                    exc_value.update(self.current_error)
                self.current_error = exc_value
                self.has_error = True
                return True
        else:
            self.stack.pop()
            self.has_error = False
        return False

    def parse(self, what=None, output=None):
        """Parse stream."""
        if what is None:  # Root item.
            output = []
            self.parse(self.root, output)
            return output[0]
        # Call the parse class
        if isinstance(what, Parser):
            inside_output = output[:]
            what(self, inside_output)
            output[:] = inside_output

        # make the what into class, and call the parse on the class
        elif isinstance(what, str):
            inside_output = [RINGToken(what)]
            if self.debug:
                print('%s%r {' % (' '*(self.depth*4), what))
                self.depth += 1
            try:
                self.parse(self.rules[what], inside_output)
            except RINGSyntaxError as exc:
                if self.debug:
                    print('%s%s' % (' '*(self.depth*4), exc))
                raise
            else:
                output.append(inside_output)
                if self.debug:
                    print('%s%r' % (' '*(self.depth*4), output))
            finally:
                if self.debug:
                    self.depth -= 1
                    print('%s} %r' % (' '*(self.depth*4), what))
        else:
            raise TypeError("Don't know how to parse type %s" % (what,))


def parse(stream, strict=False):
    if strict:
        from .Grammar import strict_grammar as grammar
    else:
        from .Grammar import enhanced_grammar as grammar
    return ParseState(grammar, stream).parse()



from warnings import warn

import numpy as np
from math import isclose

from .. Error import (
    IncompleteDataError, IncompleteDataWarning, ReadOnlyDataError,
    OutsideCorrelationError)
from .. Consts import GAS_CONSTANT as R
from .. Units import with_units
from .. ThermoChem import ThermochemBase, ThermochemRawData
from .. import yaml_io


__all__ = ['ThermochemIncomplete']


class ThermochemIncomplete(ThermochemBase):
    """
    Implement thermochemical property correlation for data missing
    heat capacity.

    Attempts to evaluate |eq_ND_H_T| or |eq_ND_S_T| will fail by raising
    :exc:`chemtk.error.IncompleteDataError` if `ND_H_ref` or `ND_S_ref`
    are respectively missing.

    If `ND_Cp_data` are missing, then attempts to evaluate |eq_ND_Cp_T|
    will also fail.  Attempts to evaluate |eq_ND_H_T| or |eq_ND_S_T| at
    temperatures other than |eq_Tref| will assume |eq_ND_Cp_T| is zero and
    will issue an :exc:`chemtk.error.IncompleteDataWarning`.

    If `ND_Cp_data` are present, quantities will be evaluated
    (where otherwise possible ) using a
    :class:`chemtk.thermochem.ThermochemRawData` correlation.
    """
    def __init__(self, ND_H_ref=None, ND_S_ref=None, ND_Cp_data={},
                 T_ref=298.15, range=None):
        """
        Initialize thermochemical property correlation for incomplete data.

        Parameters
        ----------
        ND_H_ref : float, optional
            Non-dimensional standard heat of formation |eq_ND_H_ref|
        ND_S_ref : float, optional
            Non-dimensional standard state reference entropy |eq_ND_S_ref|
        ND_Cp_data : mapping, optional
            Non-dimensional standard state heat capacity data |eq_ND_Cp_T|,
            provided as a mapping from temperature [K] to non-dimensional
            heat capacity.
        T_ref : float, optional
            Reference temperature |eq_Tref| for `ND_H_ref` and `ND_S_ref`
            (default: 298.15K).
        range : tuple(float, float), optional
            ``(lb, ub) = range`` where lb and ub are respectively the lower and
             uppers bounds of temperatures [K] for which the correlation is
             valid.
        """
        ThermochemBase.__init__(self, range=range)

        self.ND_H_ref = ND_H_ref
        self.ND_S_ref = ND_S_ref
        self.ND_Cp_data = ND_Cp_data.copy()
        self.T_ref = T_ref
        self._setup_correlation()

    def _expand_ND_Cp_data(self, ND_Cp_data):
        if not ND_Cp_data:
            return [], []
        else:
            return list(zip(*sorted(
                list(ND_Cp_data.items()), key=lambda item: item[0])))

    def _setup_correlation(self):
        if hasattr(self, '_correlation'):
            del self._correlation

        if self.ND_Cp_data:
            ND_H_ref = self.ND_H_ref if self.ND_H_ref is not None else 0.0
            ND_S_ref = self.ND_S_ref if self.ND_S_ref is not None else 0.0
            (Ts, ND_Cps) = self._expand_ND_Cp_data(self.ND_Cp_data)
            self._correlation = ThermochemRawData(
                ND_H_ref, ND_S_ref, Ts, ND_Cps, self.T_ref, self.get_range())

    def has_ND_Cp(self, T=None):
        """Return True if correlation has |eq_ND_Cp_T| data (possibly at `T`).

        Parameters
        ----------
        T : float, optional
            If `T` is specified, then check for a data point explicitly at `T`.
            Otherwise, check for the existence of any data points.
        """
        if T is None:
            return bool(self.ND_Cp_data)
        else:
            return T in self.ND_Cp_data

    def has_ND_H(self):
        """Return True if correlation has |eq_ND_H_T| data."""
        return bool(self.ND_H_ref)

    def has_ND_S(self):
        """Return True if correlation has |eq_ND_S_T| data."""
        return bool(self.ND_S_ref)

    def del_ND_Cp(self, T=None):
        """Delete |eq_ND_Cp_T| data (possibly at `T` only).

        Parameters
        ----------
        T : float, optional
            If `T` is specified, then delete data point at `T`.  Otherwise
            delete all |eq_ND_Cp_T| data.
        """
        if T is None:
            self.ND_Cp_data = None
        else:
            del self.ND_Cp_data[T]
        self._setup_correlation()

    def del_ND_H_ref(self):
        """Delete |eq_ND_H_ref| data."""
        self.ND_H_ref = None

    def del_ND_S_ref(self):
        """Delete |eq_ND_S_ref| data."""
        self.ND_S_ref = None

    def get_CpoR(self, T):
        if not self.ND_Cp_data:
            raise IncompleteDataError(
                "Cannot evaluate ND_Cp: no heat capacity data is available")
        else:
            try:
                return self._correlation.get_CpoR(T)
            except OutsideCorrelationError:
                raise IncompleteDataError(
                    "Cannot evaluate ND_Cp: no heat capacity data for T=%g"
                    % T)

    get_CpoR.__doc__ = ThermochemBase.get_CpoR.__doc__

    def get_HoRT(self, T):
        if self.ND_H_ref is None:
            raise IncompleteDataError(
                "Cannot evaluate ND_H: no enthalpy data is available")
        elif not self.ND_Cp_data:
            if T != self.T_ref:
                warn(
                    "Evaluation of ND_H_ref with (T=%g <=> T_ref=%g) will not"
                    " be corrected because heat capacity data is not"
                    " available." % (T, self.T_ref),
                    IncompleteDataWarning)
            return self.ND_H_ref
        else:
            try:
                return self._correlation.get_HoRT(T)
            except OutsideCorrelationError:
                raise IncompleteDataError(
                    "Cannot evaluate ND_H: no heat capacity data for T=%g"
                    % T)
    get_HoRT.__doc__ = ThermochemBase.get_HoRT.__doc__

    def get_SoR(self, T):
        if self.ND_S_ref is None:
            raise IncompleteDataError(
                "Cannot evaluate ND_S: no entropy data is available")
        elif not self.ND_Cp_data:
            if T != self.T_ref:
                warn(
                    "Evaluation of ND_S_ref with (T=%g <=> T_ref=%g) will not"
                    " be corrected because heat capacity data is not"
                    " available." % (T, self.T_ref),
                    IncompleteDataWarning)
            return self.ND_S_ref
        else:
            try:
                return self._correlation.get_SoR(T)
            except OutsideCorrelationError:
                raise IncompleteDataError(
                    "Cannot evaluate ND_S: no heat capacity data for T=%g"
                    % T)
    get_SoR.__doc__ = ThermochemBase.get_SoR.__doc__

    def copy(self):
        """Return a copy of this correlation.

        Returns
        -------
        thermochem : :class:`ThermochemIncomplete`
            Copy of this correlation.
        """
        return type(self)(
            self.ND_H_ref, self.ND_S_ref, self.ND_Cp_data, self.T_ref,
            self.range)

    def update(self, correlation, overwrite=False):
        """Add data contained in `correlation` into this correlation object.

        Parameters
        ----------
        correlation : :class:`ThermochemIncomplete` instance
            Specify correlation to import data from.
        overwrite : bool, optional
            True if existing data should be overwritten by data in
            `correlation`.

        Raises
        ------
        chemtk.error.ReadOnlyDataError
            If `correlation` contains data that differs from data already in
            this correlation and `overwrite` is False.
        """
        if not isinstance(correlation, type(self)):
            raise TypeError("In ThermochemIncomplete.update():",
                            " argument not instance of ThermochemIncomplete.")

        # Copy existing data first.
        data_range = self.get_range()
        T_ref = self.T_ref
        ND_H_ref = self.ND_H_ref
        ND_S_ref = self.ND_S_ref
        ND_Cp_data = self.ND_Cp_data.copy()

        # Compute updated range (as the union of the two).
        other_data_range = correlation.get_range()
        if other_data_range is not None:
            if data_range is None:
                data_range = other_data_range
            else:
                data_range = (
                    min(data_range[0], other_data_range[0]),
                    max(data_range[1], other_data_range[1]))

        # Compute updated ND_Cp_data.
        # (This is done first so the new data is available below.)
        if correlation.has_ND_Cp():
            other_ND_Cp_data = correlation.ND_Cp_data
            for T in other_ND_Cp_data:
                if(not overwrite
                   and T in self.ND_Cp_data
                   and other_ND_Cp_data[T] != ND_Cp_data[T]):
                    raise ReadOnlyDataError(
                        "In ThermochemIncomplete.update():"
                        " property ND_Cp(T=%g) already exists and differs from"
                        " new value." % T)
                ND_Cp_data[T] = other_ND_Cp_data[T]
            (Ts, ND_Cps) = self._expand_ND_Cp_data(ND_Cp_data)

        else:
            (Ts, ND_Cps) = self._expand_ND_Cp_data(ND_Cp_data)

        # Compute updated ND_H_ref and ND_S_ref.
        if correlation.has_ND_H() or correlation.has_ND_S():
            other_T_ref = correlation.T_ref
            other_ND_H_ref = correlation.ND_H_ref
            other_ND_S_ref = correlation.ND_S_ref

            # Create temporary incomplete correlation using:
            #    - Aggregated Cp_data and range
            #    - ND_H_ref, ND_S_ref and T_ref from other correlation
            # This is done to translate ND_H_ref and ND_S_ref between
            # potentially differing references temperatures.
            test_correlation = type(self)(
                other_ND_H_ref, other_ND_S_ref, ND_Cp_data, other_T_ref,
                data_range)

            # Update ND_H_ref.
            if correlation.has_ND_H():
                new_ND_H_ref = test_correlation.get_HoRT(T_ref)
                if(not overwrite
                   and ND_H_ref is not None
                   and not isclose(new_ND_H_ref, ND_H_ref, rel_tol=1e-15)):
                    raise ReadOnlyDataError(
                        "In ThermochemIncomplete.update():"
                        " property ND_H_ref already exists and differs from"
                        " new value.")
                ND_H_ref = new_ND_H_ref

            # Update ND_S_ref.
            if correlation.has_ND_S():
                new_ND_S_ref = test_correlation.get_SoR(T_ref)
                if(not overwrite
                   and ND_S_ref is not None
                   and not isclose(new_ND_S_ref, ND_S_ref, rel_tol=1e-15)):
                    raise ReadOnlyDataError(
                        "In ThermochemIncomplete.update():"
                        " property ND_S_ref already exists and differs from"
                        " new value.")
                ND_S_ref = new_ND_S_ref

        # Now store new data and update internal correlation.
        self.set_range(data_range)
        self.T_ref = T_ref
        self.ND_H_ref = ND_H_ref
        self.ND_S_ref = ND_S_ref
        self.ND_Cp_data = ND_Cp_data
        self._setup_correlation()

    @classmethod
    def yaml_construct(cls, params, context):
        T_ref = params['T_ref']
        if params.get('ND_H_ref') is not None:
            ND_H_ref = params['ND_H_ref']
        elif params.get('H_ref') is not None:
            ND_H_ref = params['H_ref']/(R*T_ref)
        else:
            ND_H_ref = None

        if params.get('ND_S_ref') is not None:
            ND_S_ref = params['ND_S_ref']
        elif params.get('S_ref') is not None:
            ND_S_ref = params['S_ref']/R
        else:
            ND_S_ref = None

        if params.get('ND_Cp_data'):
            (T_data, ND_Cp_data) = list(zip(*params['ND_Cp_data']))
            Ts = np.array([T.in_units('K') for T in T_data])
            ND_Cps = np.array(ND_Cp_data)
            ND_Cp_data = dict(list(zip(Ts, ND_Cps)))
        elif params.get('Cp_data'):
            (T_data, Cp_data) = list(zip(*params['Cp_data']))
            Ts = np.array([T.in_units('K') for T in T_data])
            ND_Cps = np.array([Cp/R for Cp in Cp_data])
            ND_Cp_data = dict(list(zip(Ts, ND_Cps)))
        else:
            ND_Cp_data = {}

        range = params.get('range')
        if range is not None:
            range = range[0].in_units('K'), range[1].in_units('K')
        T_ref = T_ref.in_units('K')

        return cls(ND_H_ref, ND_S_ref, ND_Cp_data, T_ref, range)

    def yaml_format(self, units={}):
        """Output YAML format definition of this object.

        Other Parameters
        ----------------
        units : dict
            Map containing one or more of the following keys:

                :``molar enthalpy``: units to use for enthalpy values
                :``molar entropy``: units to use for entropy values
                :``molar heat capacity``: units to use for heat capacity values
                :``temperature``: units to use for temperature values

            If any of ``molar enthalpy``, ``molar entropy``, or
            ``molar heat capacity`` is unspecified or None, then write these as
            dimensionless values.  If ``temperature`` is unspecified, then
            write temperatures in [K].
        """
        lines = []
        T_ref = with_units(self.T_ref, 'K')
        T_units = units.get('temperature', 'K')
        lines.append('T_ref: %s' % T_ref.fmt_in_units(T_units))

        if self.has_ND_H():
            H_units = units.get('molar enthalpy')
            if H_units:
                lines.append(
                    'H_ref: %s' % (R*T_ref *
                                   self.ND_H_ref).fmt_in_units(H_units))
            else:
                lines.append('ND_H_ref: %r' % self.ND_H_ref)

        if self.has_ND_S():
            S_units = units.get('molar entropy')
            if S_units:
                lines.append(
                    'S_ref: %s' % (R*self.ND_S_ref).fmt_in_units(S_units))
            else:
                lines.append('ND_S_ref: %r' % self.ND_S_ref)

        if self.has_ND_Cp():
            Cp_units = units.get('molar heat capacity')
            if Cp_units:
                lines.append('Cp_data:')
                for T in sorted(self.ND_Cp_data):
                    lines.append('    - [%s, %s]' % (
                        with_units(T, 'K').fmt_in_units(T_units),
                        (R*self.ND_Cp_data[T]).fmt_in_units(Cp_units)))
            else:
                lines.append('ND_Cp_data:')
                for T in sorted(self.ND_Cp_data):
                    lines.append('    - [%s, %r]' % (
                        with_units(T, 'K').fmt_in_units(T_units),
                        self.ND_Cp_data[T]))

        range = self.get_range()
        if range is not None:
            lines.append('range: [%s, %s]' % (
                with_units(range[0], 'K').fmt_in_units(T_units),
                with_units(range[1], 'K').fmt_in_units(T_units)))
        return '\n'.join(lines)

    _yaml_schema = """
range:
  type: tuple
  item_types: [{type: qty, kind: temperature},
               {type: qty, kind: temperature}]
  optional: true
  desc: range of valid temperatures

T_ref:
  type: qty
  kind: temperature
  default: 298.15 K
  desc: reference temperature for reference enthalpy and entropy

ND_Cp_data:
  optional: true
  type: list
  item_type:
     type: tuple
     item_types: [{type: qty, kind: temperature}, {type: float}]
  desc: set of (T, Cp(T)/R) data point pairs, where R is the gas constant

ND_H_ref:
  optional: true
  type: float
  desc: H_ref/(R*T_ref) where R is the gas constant

ND_S_ref:
  optional: true
  type: float
  desc: S_ref/R where R is the gas constant

Cp_data:
  optional: true
  type: list
  item_type:
     type: tuple
     item_types:
       - type: qty
         kind: temperature
       - type: qty
         kind: molar heat capacity
  desc: set of (T, Cp(T)) data point pairs

H_ref:
  optional: true
  type: qty
  kind: molar enthalpy
  desc: reference enthalpy

S_ref:
  optional: true
  type: qty
  kind: molar entropy
  desc: reference entropy
"""


yaml_io.register_class('ThermochemIncomplete',
                       yaml_io.parse(ThermochemIncomplete._yaml_schema),
                       ThermochemIncomplete)



from .. Error import RINGReaderError
from .. RDkitWrapper.ReactionQuery import ReactionQuery, BondForm, \
    BondIncrease, BondDecrease, BondModify, AtomTypeModify, BondBreak,\
    RadicalIncrease, RadicalDecrease, ChargeIncrease, ChargeDecrease
from .MolQueryRead import MolQueryReader
from rdkit import Chem


class ReactionQueryReader(object):
    """
    creates enhanced reaction query that is based on RING language

    from pgradd.RINGParser.Reader import Read
    from rdkit import Chem
    s = "\
    rule increaseBO{     \
    reactant r1{     \
    C labeled c1     \
    H labeled h1 single bond to c1    \
    }     \
    increase number of radical (c1)     \
    increase number of radical (h1)     \
    break bond(c1,h1) }"
    rxnquery = Read(s)
    mol = Chem.MolFromSmiles('CC')
    products = rxnquery.RunReactants(mol)
    for product in products:
    for p in product:
        print Chem.MolToSmiles(p)
    """

    def __init__(self, tree, RINGgroups=None):
        # ast = Abstract Syntax Tree
        self.tree = tree
        self.RINGgroups = RINGgroups
        self.atom_names = list()
        self.atom_belonging_mol = list()
        self.electronbalance = list()

    def ReadBondType(self, tree):
        if tree[0] == 'single':
            bondtype = Chem.BondType.SINGLE
            balance = 1
        elif tree[0] == 'double':
            bondtype = Chem.BondType.DOUBLE
            balance = 2
        elif tree[0] == 'triple':
            bondtype = Chem.BondType.TRIPLE
            balance = 3
        elif tree[0] == 'quadruple':
            bondtype = Chem.BondType.QUADRUPLE
            balance = 4
        elif tree[0] == 'aromatic':
            bondtype = Chem.BondType.AROMATIC
            balance = 1.5
        elif tree[0] == 'partial':
            bondtype = Chem.BondType.DATIVE
            balance = 0
        else:
            raise RINGReaderError("BondType: Unsupported bond type: '"
                                  + tree[0] + "'")
        return bondtype, balance

    def ReadBondForm(self, tree, reactionquery):
        i = 0
        if tree[i][0] == 'BondType':
            bondtype, balance = self.ReadBondType(tree[i][1:])
            i += 1
        else:
            bondtype = Chem.BondType.SINGLE
            balance = 1
        try:
            idx1 = self.atom_names.index(tree[i][1])
            i += 1
        except Exception:
            raise RINGReaderError("BondForm: Unrecognized atom label: '"
                                  + tree[i][1] + "'")
        try:
            idx2 = self.atom_names.index(tree[i][1])
        except Exception:
            raise RINGReaderError("BondForm: Unrecognized atom label: '"
                                  + tree[i][1] + "'")
        self.electronbalance[idx1] += -balance
        self.electronbalance[idx2] += -balance
        reactionquery.transformations.append(BondForm(idx1, idx2, bondtype))

    def ReadBondBreak(self, tree, reactionquery):
        i = 0
        if tree[i][0] == 'BondType':
            bondtype, balance = self.ReadBondType(tree[i][1:])
            i += 1
        else:
            bondtype = Chem.BondType.SINGLE
            balance = +1

        try:
            atom_name1 = tree[i][1]
            idx1 = self.atom_names.index(atom_name1)
            reactant_name1 = self.atom_belonging_mol[idx1]
            idx1_in_query = reactionquery.reactantquery[reactant_name1].\
                atom_names.index(atom_name1)
            i += 1
        except Exception:
            raise RINGReaderError("BondBreak: Unrecognized atom label: '"
                                  + tree[i][1] + "'")
        try:
            atom_name2 = tree[i][1]
            idx2 = self.atom_names.index(atom_name2)
            reactant_name2 = self.atom_belonging_mol[idx2]
            idx2_in_query = reactionquery.reactantquery[reactant_name2].\
                atom_names.index(atom_name2)
        except Exception:
            raise RINGReaderError("BondBreak: Unrecognized atom label: '"
                                  + tree[i][1] + "'")
        if reactant_name1 != reactant_name2:
            raise RINGReaderError("BondBreak: Two supplied atom labels aren't",
                                  "in the same molecule")
        bond = reactionquery.reactantquery[reactant_name1].mol.\
            GetBondBetweenAtoms(idx1_in_query, idx2_in_query)
        if not bond:
            raise RINGReaderError("BondBreak: No bond between two atom ("
                                  + atom_name1 + "," + atom_name2 + ")")
        elif bond.GetBondType().__str__() == 'UNSPECIFIED':
            raise RINGReaderError("BondBreak: Break of unspecified bond",
                                  "is not supported ")
        elif bond.GetBondType() != bondtype:
            raise RINGReaderError("BondBreak: Bond mismatch between two",
                                  "atom (" + atom_name1 + "," + atom_name2
                                  + "). (Bond in mol:"
                                  + bond.GetBondType().__str__() +
                                  " ,bond broken:" + bondtype.__str__())

        self.electronbalance[idx1] += balance
        self.electronbalance[idx2] += balance
        reactionquery.transformations.append(BondBreak(idx1, idx2))

    def ReadBondModify(self, tree, reactionquery):
        try:
            atom_name1 = tree[0][1]
            idx1 = self.atom_names.index(atom_name1)
            reactant_name1 = self.atom_belonging_mol[idx1]
            idx1_in_query = reactionquery.reactantquery[reactant_name1].\
                atom_names.index(atom_name1)
        except Exception:
            raise RINGReaderError("BondModify: Unrecognized atom label: '"
                                  + tree[0][1] + "'")
        try:
            atom_name2 = tree[1][1]
            idx2 = self.atom_names.index(atom_name2)
            reactant_name2 = self.atom_belonging_mol[idx2]
            idx2_in_query = reactionquery.reactantquery[reactant_name2].\
                atom_names.index(atom_name2)
        except Exception:
            raise RINGReaderError("BondModify: Unrecognized atom label: '"
                                  + tree[1][1] + "'")
        if reactant_name1 != reactant_name2:
            raise RINGReaderError("BondModify: Two supplied atom labels",
                                  "aren't in the same molecule")
        bond = reactionquery.reactantquery[reactant_name1].\
            mol.GetBondBetweenAtoms(idx1_in_query, idx2_in_query)
        if not bond:
            raise RINGReaderError("BondModify: No bond between two atom ("
                                  + atom_name1 + "," + atom_name2 + ")")
        elif bond.GetBondType().__str__() == 'UNSPECIFIED':
            raise RINGReaderError("BondModify: BondModify: modification of",
                                  "unspecified bond is not supported ")
        existingbond = bond.GetBondType().__str__()
        if existingbond == 'SINGLE':
            balance1 = 1
        elif existingbond == 'DOUBLE':
            balance1 = 2
        elif existingbond == 'TRIPLE':
            balance1 = 3
        elif existingbond == 'QUADRUPLE':
            balance1 = 4
        elif existingbond == 'AROMATIC':
            balance1 = 1.5
        elif existingbond == 'DATIVE':
            balance1 = 0
        else:
            raise RINGReaderError("BondModify: Unsupported bond type for",
                                  "modifiation: '"+existingbond+"'")
        bondtype, balance = self.ReadBondType(tree[2][1:])
        self.electronbalance[idx1] -= balance - balance1
        self.electronbalance[idx2] -= balance - balance1
        reactionquery.transformations.append(BondModify(idx1, idx2, bondtype))

    def ReadBondIncrease(self, tree, reactionquery):
        try:
            idx1 = self.atom_names.index(tree[0][1])
        except Exception:
            raise RINGReaderError("BondIncrease: Unrecognized atom label: '"
                                  + tree[0][1] + "'")
        try:
            idx2 = self.atom_names.index(tree[1][1])
        except Exception:
            raise RINGReaderError("BondIncrease: Unrecognized atom label: '"
                                  + tree[1][1] + "'")

        self.electronbalance[idx1] += -1
        self.electronbalance[idx2] += -1
        reactionquery.transformations.append(BondIncrease(idx1, idx2))

    def ReadBondDecrease(self, tree, reactionquery):
        try:
            idx1 = self.atom_names.index(tree[0][1])
        except Exception:
            raise RINGReaderError("BondDecrease: Unrecognized atom label: '"
                                  + tree[0][1] + "'")
        try:
            idx2 = self.atom_names.index(tree[1][1])
        except Exception:
            raise RINGReaderError("BondDecrease: Unrecognized atom label: '"
                                  + tree[1][1] + "'")

        self.electronbalance[idx1] += 1
        self.electronbalance[idx2] += 1
        reactionquery.transformations.append(BondDecrease(idx1, idx2))

    def ReadAtomSuffix(self, tree):
        # '+','-','.',':','+.','-.','*'
        if tree[0] == '+.':
            radical = 1
            charge = 1
            valence = 0
        elif tree[0] == '-.':
            radical = 1
            charge = -1
            valence = 0
        elif tree[0] == '+':
            radical = 0
            charge = 1
            valence = 0
        elif tree[0] == '-':
            radical = 0
            charge = -1
            valence = 0
        elif tree[0] == '.':
            radical = 1
            charge = 0
            valence = 0
        elif tree[0] == ':':
            radical = 2
            charge = 0
            valence = 0
        elif tree[0] == '*':
            radical = 0
            valence = 1
            charge = 1
        else:
            s = "AtomSuffic: Unsupported atom suffic: '" + tree[0] + "'"
            raise NotImplementedError(s)
        return radical, charge, valence

    def ReadAtomType(self, tree):
        assert tree[0][0] == 'Symbols'
        symbol = tree[0][1][0]

        if len(tree) > 1:
            assert tree[1][0] == 'AtomSuffix'
            radical, charge, valence = self.ReadAtomSuffix(tree[1][1:])
        return symbol, radical, charge, valence

    def ReadAtomLabel(self, tree, reactionquery):
        try:
            atom_name = tree[0]
            idx = self.atom_names.index(atom_name)
            reactant_name = self.atom_belonging_mol[idx]
            idx_in_query = reactionquery.reactantquery[reactant_name].\
                atom_names.index(atom_name)
            atom = reactionquery.reactantquery[reactant_name].\
                mol.GetAtomWithIdx(idx_in_query)
        except Exception:
            raise RINGReaderError("AtomLabel: Unrecognized atom label: '"
                                  + tree[0] + "'")
        return atom_name, idx, reactant_name, idx_in_query, atom

    def ReadAtomTypeModify(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        _, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:], reactionquery)
        assert tree[1][0] == 'AtomType'
        symbol, radical, charge, valence = self.ReadAtomType(tree[1][1:])
        if atom.GetSymbol() != symbol:
            raise NotImplementedError("AtomTypeModify: Atom Label change",
                                      "not supported")

        self.electronbalance[idx] -= radical - atom.GetNumRadicalElectrons()
        self.electronbalance[idx] += charge - atom.GetFormalCharge()
        reactionquery.transformations.append(AtomTypeModify(idx,
                                                            radical,
                                                            charge,
                                                            valence))

    def ReadRadicalModify(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        _, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:], reactionquery)
        radical = tree[1]
        if radical < 0:
            raise RINGReaderError("RadicalModify: Number of radical",
                                  "electrons cannot be below 0")
        self.electronbalance[idx] -= radical - atom.GetNumRadicalElectrons()
        reactionquery.transformations.append(AtomTypeModify(idx,
                                                            radical,
                                                            0, 0))

    def ReadRadicalIncrease(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        _, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:], reactionquery)

        self.electronbalance[idx] -= 1
        reactionquery.transformations.append(RadicalIncrease(idx))

    def ReadRadicalDecrease(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        atom_name, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:],
                                                        reactionquery)
        self.electronbalance[idx] += 1
        reactionquery.transformations.append(RadicalDecrease(idx))

    def ReadChargeIncrease(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        _, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:], reactionquery)

        self.electronbalance[idx] -= 1
        reactionquery.transformations.append(ChargeIncrease(idx))

    def ReadChargeDecrease(self, tree, reactionquery):
        assert tree[0][0] == 'AtomLabel'
        atom_name, idx, _, _, atom = self.ReadAtomLabel(tree[0][1:],
                                                        reactionquery)
        self.electronbalance[idx] += 1
        reactionquery.transformations.append(ChargeDecrease(idx))

    def ReadConnectivityChange(self, tree, reactionquery):
        if tree[0][0] == 'BondForm':
            self.ReadBondForm(tree[0][1:], reactionquery)
        elif tree[0][0] == 'BondBreak':
            self.ReadBondBreak(tree[0][1:], reactionquery)
        elif tree[0][0] == 'BondModify':
            self.ReadBondModify(tree[0][1:], reactionquery)
        elif tree[0][0] == 'BondIncrease':
            self.ReadBondIncrease(tree[0][1:], reactionquery)
        elif tree[0][0] == 'BondDecrease':
            self.ReadBondDecrease(tree[0][1:], reactionquery)
        elif tree[0][0] == 'AtomTypeModify':
            self.ReadAtomTypeModify(tree[0][1:], reactionquery)
        elif tree[0][0] == 'RadicalModify':
            self.ReadRadicalModify(tree[0][1:], reactionquery)
        elif tree[0][0] == 'RadicalIncrease':
            self.ReadRadicalIncrease(tree[0][1:], reactionquery)
        elif tree[0][0] == 'RadicalDecrease':
            self.ReadRadicalDecrease(tree[0][1:], reactionquery)
        elif tree[0][0] == 'ChargeIncrease':
            self.ReadChargeIncrease(tree[0][1:], reactionquery)
        elif tree[0][0] == 'ChargeDecrease':
            self.ReadChargeDecrease(tree[0][1:], reactionquery)

    def ReadTransformationChain(self, tree, reactionquery):
        assert tree[0][0] == 'ConnectivityChange'
        self.ReadConnectivityChange(tree[0][1:], reactionquery)
        if 1 < len(tree):
            assert tree[1][0] == 'TransformationChain'
            self.ReadTransformationChain(tree[1][1:], reactionquery)

    def LabelMapping(self, tree, labelmapping=None):
        if not labelmapping:
            labelmapping = dict()
        assert tree[0][0] == 'AtomLabel'
        assert tree[1][0] == 'AtomLabel'
        labelmapping[tree[0][1]] = tree[1][1]
        if 2 < len(tree):
            assert tree[2][0] == 'LabelMapping'
            labelmapping = self.LabelMapping(tree[2][1:], labelmapping)
        return labelmapping

    def ReadReactantGroup(self, tree, reactionquery):
        assert tree[0][0] == 'ReactantName'
        assert tree[1][0] == 'GroupName'
        assert tree[2][0] == 'LabelMapping'
        labelmapping = self.LabelMapping(tree[2][1:])
        if tree[1][1] not in self.RINGgroups:
            raise RINGReaderError("ReactantGroup: Unrecognized group name:'"
                                  + tree[1][1] + "'")

        if len(labelmapping) != len(self.RINGgroups[tree[1][1]]):
            raise RINGReaderError("ReactantGroup: Label mapping length",
                                  "doesn''t match length of atoms in group :'"
                                  + tree[1][1] + "'")
        reactionquery.reactantquery[tree[0][1]] = self.RINGgroups[tree[1][1]]
        reactionquery.reactantquery[tree[0][1]].name = tree[0][1]
        for i in range(0, len(reactionquery.reactantquery[tree[0][1]].
                              atom_names)):
            try:
                reactionquery.reactantquery[tree[0][1]].atom_names[i] = \
                    labelmapping[reactionquery.reactantquery[tree[0][1]].
                                 atom_names[i]]
            except KeyError:
                s = 'Unrecognized label '
                s += reactionquery.reactantquery[tree[0][1]].atom_names[i]
                s += ' in reactant ' +\
                    reactionquery.reactantquery[tree[0][1]].name
                s += ' for duplication opreation for group ' + tree[1][1]
                raise RINGReaderError(s)
        self.atom_names += reactionquery.reactantquery[tree[0][1]].atom_names
        self.electronbalance +=\
            [0]*len(reactionquery.reactantquery[tree[0][1]].atom_names)
        self.atom_belonging_mol +=\
            [0]*len(reactionquery.reactantquery[tree[0][1]].name)

    def ReadDuplicates(self, tree, reactionquery):
        assert tree[0][0] == 'ReactantName'
        assert tree[1][0] == 'ReactantName'
        assert tree[2][0] == 'LabelMapping'
        labelmapping = self.LabelMapping(tree[2][1:])
        if len(labelmapping) != len(reactionquery.reactantquery[tree[1][1]].
                                    atom_names):
            raise RINGReaderError('ReadDuplicates: Labelmapping length',
                                  "doesn''t match length of fragment",
                                  'duplicated')
        reactionquery.reactantquery[tree[0][1]] = reactionquery.\
            reactantquery[tree[1][1]]
        reactionquery.reactantquery[tree[0][1]].name = tree[0][1]
        for i in range(0, len(reactionquery.reactantquery[tree[0][1]].
                              atom_names)):
            try:
                reactionquery.reactantquery[tree[0][1]].atom_names[i] = \
                    labelmapping[reactionquery.reactantquery[tree[0][0]].
                                 atom_names[i]]
            except KeyError:
                s = 'Unrecognized label '
                s += reactionquery.reactantquery[tree[0][1]].atom_names[i]
                s += ' in reactant ' +\
                    reactionquery.reactantquery[tree[1][1]].name
                s += ' for duplication opreation for reactant ' + tree[0][1]
                raise RINGReaderError(s)
        self.atom_names += reactionquery.reactantquery[tree[0][1]].atom_names
        self.electronbalance +=\
            [0]*len(reactionquery.reactantquery[tree[0][1]].atom_names)
        self.atom_belonging_mol += reactionquery.reactantquery[tree[0][1]].\
            name*len(reactionquery.reactantquery[tree[0][1]].atom_names)

    def ReadReactants(self, tree, reactionquery):
        assert tree[0][0] in ['ReactantQuery', 'ReactantGroup', 'Duplicates']
        if tree[0][0] == 'ReactantQuery':
            molquery = MolQueryReader(tree[0][1:]).Read()
            reactionquery.AppendReactantQuery(molquery)
            self.atom_names += molquery.atom_names
            self.electronbalance += [0]*len(molquery.atom_names)
            self.atom_belonging_mol += [molquery.name]*len(molquery.atom_names)
        elif tree[0][0] == 'ReactantGroup':
            self.ReadReactantGroup(tree[1][1:], reactionquery)
        elif tree[0][0] == 'Duplicates':
            self.ReadDuplicates(tree[1][1:], reactionquery)
        if len(tree) == 2:
            self.ReadReactants(tree[1][1:], reactionquery)

    def Read(self):
        # initialize Enhanced ReactionQuery
        reactionquery = ReactionQuery()
        # fragment name
        i = 0
        assert self.tree[i][0] == 'ReactionName'
        reactionquery.name = self.tree[i][1]
        i += 1
        # start building reactionquery
        assert self.tree[i][0] == 'Reactants'
        self.ReadReactants(self.tree[i][1:], reactionquery)
        i += 1
        if self.tree[i][0] == 'Constraints':
            raise NotImplementedError("Constraints not supported yet")
            self.ReadConstraints(self.tree[i][1:], reactionquery)
            i += 1
        # graph transformation is done here.
        assert self.tree[i][0] == 'TransformationChain'
        self.ReadTransformationChain(self.tree[i][1:], reactionquery)
        # error check
        s = str()
        for i in range(0, len(self.electronbalance)):
            if self.electronbalance[i] != 0:
                s += "Electron balance mismatch for atom label: '" +\
                    self.atom_names[i] + "' by ("\
                    + str(self.electronbalance[i]) + ")\n"
        if s:
            raise RINGReaderError(s)
        # return
        return reactionquery



import yaml

from .common import string


__all__ = ['parse', 'YAMLTaggedValue']


class YAMLTaggedValue(object):
    def __init__(self, tag, value):
        self.tag = tag
        self.value = value

    def __str__(self):
        return '<%s>: %s' % (self.tag, self.value)

    def __repr__(self):
        return 'YAMLTaggedValue(%r, %r)' % (self.tag, self.value)


def _construct_str(loader, node):
    return string(loader.construct_scalar(node))


def _construct_seq(loader, node):
    result = loader.construct_sequence(node)
    return result


def _construct_map(loader, node):
    result = dict([(tuple(key), value) if isinstance(key, list)
                  else (key, value)
                  for (key, value) in loader.construct_pairs(node, deep=True)])
    return result


def _construct_tagged(loader, tag, node):
    if isinstance(node, yaml.MappingNode):
        value = _construct_map(loader, node)
    elif isinstance(node, yaml.SequenceNode):
        value = _construct_seq(loader, node)
    elif isinstance(node, yaml.ScalarNode):
        value = loader.construct_scalar(node)
    else:
        raise TypeError('Unknown node type: %s' % type(node))
    return YAMLTaggedValue(tag, value)


try:
    # Use the class that's atop libyaml for speed...
    class Loader(yaml.CSafeLoader):
        pass
except AttributeError:
    # ...and fall back on the Python implementation otherwise:
    class Loader(yaml.SafeLoader):
        pass

# Note: other possible tags:
#   'tag:yaml.org,2002:int', 'tag:yaml.org,2002:float',
#   'tag:yaml.org,2002:bool', tag:yaml.org,2002:null',
#   and 'tag:yaml.org,2002:seq'
Loader.add_constructor('tag:yaml.org,2002:str', _construct_str)
Loader.add_constructor('tag:yaml.org,2002:map', _construct_map)

# Captures all things with an explicit tag.
Loader.add_multi_constructor('', _construct_tagged)

# Might look into these later?
# VGToolsLoader.add_multi_constructor('!', VGToolsLoader.constructor)
# VGToolsLoader.add_constructor('!!seq', VGToolsLoader.seq_constructor)


# We want YAML 1.2 support but the Python YAML library only supports YAML 1.1.
# YAML 1.1 has lots of implicit resolvers that try to automatically convert
# too many things.  We address this by replacing the implicit_resolvers with
# only those that are explicitly in the YAML 1.2 core set).
YAML12_core_implicit_tags = [
    'tag:yaml.org,2002:' + tag for tag in
    ['str', 'seq', 'map', 'null', 'bool', 'int', 'float']]
for first in Loader.yaml_implicit_resolvers:
    Loader.yaml_implicit_resolvers[first] = [
        (tag, regexp)
        for (tag, regexp) in Loader.yaml_implicit_resolvers[first]
        if tag in YAML12_core_implicit_tags]


def parse(yaml_input):
    """Parse data in YAML format to an abstract tree."""
    return yaml.load(yaml_input, Loader=Loader)



"""
====================================
Thermochemical property correlations
====================================

This module provides a variety of different correlations used to represent
thermochemical properties (in non-dimensional form), i.e. heat capacity
|eq_ND_Cp_T|, enthalpy of formation |eq_ND_H_T|, and absolute |eq_ND_S_T| all
as functions of temperature |eq_T|.

This module provides a group-additivity *property set* for estimation of
thermochemical properties.  Upon importing this module, the ``thermochem``
property set is registered for use within
:class:`chemtk.groupadd.GroupLibrary` instances.  Group contributions may also
be computed by fitting to thermochemical data via
:class:`ThermochemGroupFittingSet()`.objects.

---------
Reference
---------

.. autodoc puts stuff extracted from docstrings here.
"""

# TODO: Test ThermochemQuartic code and add to correlations.
from .base import *
from .raw_data import *
from .incomplete import *
from .group_data import *

__all__ = sum((getattr(globals()[module], '__all__') for module in [
    'base', 'raw_data', 'incomplete', 'group_data']), [])

del base, raw_data, incomplete, group_data



from ..Error import MolQueryError
from rdkit import Chem
from collections import defaultdict
import operator as op
"""
Written by Geun Ho Gu 8/4/2016
MolQuery class is equipped with enhanced query comparison over
RDkit query comparison.
This module uses a graph description language called R.I.N.G.
Rangarajan, S., Kaminski, T., Van Wyk, E., Bhan A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Algorithms of RING", Computers and Chemical Engineering 64 (2014) 124,
    10.1016/j.compchemeng.2014.02.007
Rangarajan, S., Bhan, A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Descrpition of RING", Computers and Chemical Engineering 45 (2012) 114,
    10.1016/j.compchemeng.2012.06.008
"""
ops = {'>': op.gt,
       '<': op.lt,
       '>=': op.ge,
       '<=': op.le,
       '=': op.eq}
digit = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '-', 'E', 'e']


class ConstraintNumber(object):
    def __init__(self, s):
        if type(s) == list and len(s) == 2:  # given a list ['>=',2]
            self.operator = s[0]
            self.n = s[1]
        elif type(s) == list and len(s) == 1:
            self.operator = '='
            self.n = s[0]
        elif type(s) == str and s[0].isdigit():  # given a string number [3]
            self.operator = '='
            self.n = int(s)
        elif type(s) == int:
            self.operator = '='
            self.n = s
        else:  # given a string '>= 3'
            i = 1
            while not s[i] in digit and s[i] != ' ':
                i += 1
            self.operator = s[:i]
            self.n = int(s[i:])

    def __call__(self, inp):
        return ops[self.operator](inp, self.n)

    def __str__(self):
        return '<ConstraintNumber(%s%s)>' % (self.operator, self.n)

    def __repr__(self):
        return 'ConstraintNumber(%s%s)' % (self.operator, self.n)


class BondConstraint(object):
    def __init__(self, bondquery):
        assert isinstance(bondquery, BondQuery)
        self.bondquery = bondquery

    def __call__(self, idx1, idx2, mol):
        try:
            bond = mol.GetBondBetweenAtoms(idx1, idx2)
        except Exception:
            raise MolQueryError('There is no bond between atom'
                                + idx1+' and atom'+idx2)
        return self.bondquery(bond)

    def __str__(self):
        return '<BondConstraint(%s)>' % (self.bondquery)

    def __repr__(self):
        return 'BondConstraint(%s)' % (self.bondquery)

# BondQuery is not combined with BondConstraint because
# Bondquery is used elsewhere


class BondQuery(object):
    def __init__(self, RINGbondname):
        if RINGbondname not in ['single', 'double', 'triple',
                                'quadruple', 'ring', 'nonring',
                                'aromatic', 'any', 'strong', 'partial']:
            raise NotImplementedError("Unsupported bond type '"
                                      + RINGbondname + "'")
        self.RINGbondname = RINGbondname

    def __call__(self, rdkitbond):
        assert isinstance(rdkitbond, Chem.Bond)
        if self.RINGbondname == 'single':
            if rdkitbond.GetBondType() == Chem.BondType.SINGLE:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'double':
            if rdkitbond.GetBondType() == Chem.BondType.DOUBLE:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'triple':
            if rdkitbond.GetBondType() == Chem.BondType.TRIPLE:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'quadruple':
            if rdkitbond.GetBondType() == Chem.BondType.QUADRUPLE:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'ring':
            if rdkitbond.IsInRing():
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'nonring':
            if not rdkitbond.IsInRing():
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'aromatic':
            if rdkitbond.GetBondType() == Chem.BondType.AROMATIC:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'any':
            return True
        elif self.RINGbondname == 'strong':
            if rdkitbond.GetBondType() in [Chem.BondType.DOUBLE,
                                           Chem.BondType.TRIPLE,
                                           Chem.BondType.QUADRUPLE,
                                           Chem.BondType.AROMATIC]:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        elif self.RINGbondname == 'partial':
            if rdkitbond.GetBondType() in [Chem.BondType.DATIVE,
                                           Chem.BondType.OTHER,
                                           Chem.BondType.ZERO]:
                return True
            else:
                raise MolQueryError('BondQuery: False.')
        else:
            raise NotImplementedError("Unsupported bond type '"
                                      + rdkitbond.GetBondType().name+"'")

    def __str__(self):
        return '<BondQuery(%s)>' % (self.RINGbondname)

    def __repr__(self):
        return 'BondQuery(%s)' % (self.RINGbondname)


class DoubleBondStereoConstraint(object):
    def __init__(self, negate, stereobondtype):
        self.negate = negate
        self.stereobondtype = stereobondtype

    def __call__(self, idx1, idx2, idx3, idx4, mol):
        double_bond = mol.GetBondBetweenAtoms(idx3, idx4)
        stereo = double_bond.GetStereo()
        # Chem.rdchem.BondStereo.STEREOZ = cis
        # Chem.rdchem.BondStereo.STEREOE = trans
        # Chem.rdchem.BondStereo.STEREONONE = Not specified
        # non-type check
        if stereo == Chem.rdchem.BondStereo.STEREONONE:
            if self.stereobondtype == stereo and not self.negate:
                return True
            elif self.stereobondtype != stereo and self.negate:
                return True
            else:
                raise MolQueryError('DoubleBondStereoConstraint: False.')
        # Cis trans. Here I use a slightly clever algorithm.
        # I look at the intersectio between the one specified by rdkit.
        nmatch = len(set(double_bond.GetStereoAtoms()).
                     intersection([idx1, idx2]))
        if nmatch == 2 or nmatch == 0:

            if self.stereobondtype == stereo and not self.negate:
                return True
            elif self.stereobondtype != stereo and self.negate:
                return True
            else:
                raise MolQueryError('DoubleBondStereoConstraint: False.')
        elif nmatch == 1:
            # if there is only one match, stereo is opposite.
            if stereo == Chem.rdchem.BondStereo.STEREOZ:
                stereo = Chem.rdchem.BondStereo.STEREOE
            elif stereo == Chem.rdchem.BondStereo.STEREOE:
                stereo = Chem.rdchem.BondStereo.STEREOZ
            if self.stereobondtype == stereo and not self.negate:
                return True
            elif self.stereobondtype != stereo and self.negate:
                return True
            else:
                raise MolQueryError('DoubleBondStereoConstraint: False.')

    def __str__(self):
        return '<DoubleBondStereoConstraint(%s%s)>' % (self.negate,
                                                       self.stereobondtype)

    def __repr__(self):
        return 'DoubleBondStereoConstraint(%s%s)' % (self.negate,
                                                     self.stereobondtype)


class AtomConstraint(object):
    pass


class AtomRadical(AtomConstraint):
    def __init__(self, negate, CN):
        self.negate = negate
        self.CN = CN

    def __call__(self, atom):
        if self.negate and self.CN(atom.GetNumRadicalElectrons()):
            raise MolQueryError('AtomRadical: False.')
        elif not self.negate and not self.CN(atom.GetNumRadicalElectrons()):
            raise MolQueryError('AtomRadical: False.')

    def __repr__(self):
        return 'AtomRadical()'


class AtomIsInRing(AtomConstraint):
    def __init__(self, negate):
        self.negate = negate

    def __call__(self, atom):
        if self.negate and atom.IsInRing():
            raise MolQueryError('AtomIsInRing: False.')
        elif not self.negate and not atom.IsInRing():
            raise MolQueryError('AtomIsInRing: False.')

    def __repr__(self):
        return 'AtomIsInRing()'


class AtomIsAromatic(AtomConstraint):
    def __init__(self, negate):
        self.negate = negate

    def __call__(self, atom):
        if self.negate and atom.GetIsAromatic():
            raise MolQueryError('AtomIsAromatic: False.')
        elif not self.negate and not atom.GetIsAromatic():
            raise MolQueryError('AtomIsAromatic: False.')

    def __repr__(self):
        return 'AtomAromatic()'


class AtomIsAllylic(AtomConstraint):
    def __init__(self, negate):
        self.negate = negate

    def __call__(self, atom):
        yes = False
        for bond in atom.GetBonds():
            if bond.GetBondType() == Chem.BondType.DOUBLE:
                yes = True
                break
        if self.negate and yes:
            raise MolQueryError('AtomAllylic: False.')
        elif not self.negate and not yes:
            raise MolQueryError('AtomAllylic: False.')

    def __repr__(self):
        return 'AtomAllylic()'


class AtomRing(AtomConstraint):
    def __init__(self, negate, ring_sizeCN):
        self.negate = negate
        self.ring_sizeCN = ring_sizeCN

    def __call__(self, atom):
        if self.negate:
            rings = atom.GetOwningMol().GetRingInfo().AtomRings()
            for ring in rings:
                if atom.GetIdx() in ring:
                    if self.ring_sizeCN(len(ring)):
                        raise MolQueryError('AtomRing: False.')
        else:
            if not atom.IsInRing():
                raise MolQueryError('AtomRing: False. There is no ring.')
            else:
                rings = atom.GetOwningMol().GetRingInfo().AtomRings()
                for ring in rings:
                    if atom.GetIdx() in ring:
                        if self.ring_sizeCN(len(ring)):
                            return
                raise MolQueryError('AtomRing: False.')

    def __repr__(self):
        return 'AtomRing()'


class AtomNRing(AtomConstraint):
    def __init__(self, negate, NringCN):
        self.negate = negate
        self.NringCN = NringCN

    def __call__(self, atom):
        # get number of rings
        rings = atom.GetOwningMol().GetRingInfo().AtomRings()
        n = 0
        for ring in rings:
            for ring_atom_idx in ring:
                if atom.GetIdx() == ring_atom_idx:
                    n += 1
        if self.negate and self.NringCN(n):
            raise MolQueryError('AtomNRing: False.')
        elif not self.NringCN(n):
            raise MolQueryError('AtomNRing: False.')

    def __repr__(self):
        return 'AtomNRing()'


class AtomConnectivityAtom(AtomConstraint):
    def __init__(self, negate, CN, connected, bondquery, constraints):
        assert isinstance(negate, bool)
        assert isinstance(CN, ConstraintNumber)
        self.negate = negate
        self.ConstraintNumber = CN
        self.connected = connected
        if isinstance(connected, str):
            NotImplementedError("GroupName in AtomConstraint Not supported")
        self.bondquery = bondquery
        self.constraints = constraints

    def __call__(self, atom):
        count = 0
        for bond in atom.GetBonds():
            # atom check
            connected_atom = bond.GetOtherAtom(atom)
            if not self.connected.Match(connected_atom):
                continue
            try:
                for constraint in self.constraints:
                    constraint(connected_atom)
            except Exception:
                continue
            # bond check
            try:
                if self.bondquery(bond):
                    count += 1
            except Exception:
                continue
        if self.negate:
            if self.ConstraintNumber(count):
                raise MolQueryError('AtomConnectivity: False. (Negate)',
                                    'Matching bond: %s. Constraint:%s.'
                                    % (count, self.ConstraintNumber))
        else:
            if not self.ConstraintNumber(count):
                raise MolQueryError('AtomConnectivity: False. Matching'
                                    'bond: %s. Constraint:%s.'
                                    % (count, self.ConstraintNumber))

    def __repr__(self):
        return 'AtomConnectivityAtom()'


class AtomConnectivityGroup(AtomConstraint):
    def __init__(self, negate, CN, connected, bondquery):
        assert isinstance(negate, bool)
        assert isinstance(CN, ConstraintNumber)
        self.negate = negate
        self.ConstraintNumber = CN
        self.connected = connected
        if isinstance(connected, str):
            NotImplementedError("GroupName in AtomConstraint Not supported")
        self.bondquery = bondquery

    def __call__(self, atom):
        count = 0
        # Do mol match first.
        mol = atom.GetOwningMol()
        for match_index in mol.GetSubstructMatches(self.connected):
            bond = mol.GetBondBetweenAtoms(atom.GetIdx(), match_index[0])
            try:
                if bond and self.bondquery(bond):
                    count += 1
            except Exception:
                continue
        if self.negate:
            if self.ConstraintNumber(count):
                raise MolQueryError('AtomConnectivity: False. (Negate)',
                                    'Matching bond: %s. Constraint:%s.'
                                    % (count, self.ConstraintNumber))
        else:
            if not self.ConstraintNumber(count):
                raise MolQueryError('AtomConnectivity: False. Matching bond:',
                                    '%s. Constraint:%s.'
                                    % (count, self.ConstraintNumber))

    def __repr__(self):
        return 'AtomConnectivityAtom()'


class MolConstraint(object):
    pass


class MolCharge(MolConstraint):
    def __init__(self, CN):
        assert isinstance(CN, ConstraintNumber)
        self.ConstraintNumber = CN

    def __call__(self, mol):
        total_charge = 0
        for atom in mol.GetAtoms():
            total_charge += atom.GetFormalCharge()
        if self.ConstraintNumber(total_charge):
            return self.ConstraintNumber(total_charge)
        else:
            raise MolQueryError('MolCharge: False. Total Charge: %s.',
                                'Constraint:%s.'
                                % (total_charge, self.ConstraintNumber))

    def __repr__(self):
        return 'MolCharge(%r)' % self.n


class MolAromatic(MolConstraint):
    def __call__(self, mol):
        if not mol.GetAromaticAtoms():
            raise MolQueryError("MolAromatic: mol not aromatic")

    def __repr__(self):
        return 'MolAromatic()'


class MolOlefinic(MolConstraint):
    def __call__(self, mol):
        if not mol.HasSubstructMatch(Chem.MolFromSmiles('C=C')):
            raise MolQueryError("MolOlefinic: No double bond")

    def __repr__(self):
        return 'MolOlefinic()'


class MolParaffinic(MolConstraint):
    def __call__(self, mol):
        if mol.HasSubstructMatch(Chem.MolFromSmiles('C=C')):
            raise MolQueryError("MolParaffinic: There is double bond")

    def __repr__(self):
        return 'MolParaffinic()'


class MolCyclic(MolConstraint):
    def __call__(self, mol):
        if not mol.GetRingInfo().NumRings():
            raise MolQueryError("MolCyclic: Mol isn't cyclic")

    def __repr__(self):
        return 'MolCyclic()'


class MolLinear(MolConstraint):
    def __call__(self, mol):
        if mol.GetRingInfo().NumRings():
            raise MolQueryError("MolLinear: Mol is cyclic")

    def __repr__(self):
        return 'MolCyclic()'


class MolQuery(object):
    """Goes through constraint and check for match"""
    def __init__(self):
        self.mol = Chem.RWMol(Chem.Mol())
        self.atom_names = list()
        self.mol_constraints = list()
        self.atom_constraints = defaultdict(list)
        """
        atom_constraints = { %atomidx% : [constraint1,constrain2],...}
        """
        self.bond_constraints = list()
        """
        bond_constraints = [[atomqueryidx1,atomqueryidx2,constraint],...]
        """
        self.double_bond_stereo_constraints = list()

    def __repr__(self):
        s = "%s(" % (self.__class__.__name__)
        if hasattr(self, 'name'):
            s += "name='%s', " % (self.name)
        s += "smiles='%s, " % Chem.MolToSmiles(self.mol)
        nconstraint = len(self.mol_constraints)
        for constraint in self.atom_constraints:
            nconstraint += len(self.atom_constraints[constraint])
        s += "NumberConstraints=('%s')" % (nconstraint)
        s += ")"
        return s

    def AppendMolConstraint(self, constraint):
        assert isinstance(constraint, MolConstraint)
        self.mol_constraints.append(constraint)

    def AppendAtomConstraint(self, constraint, idx_or_atomname):
        assert isinstance(constraint, AtomConstraint)
        if isinstance(idx_or_atomname, str):
            try:
                idx = self.atom_names.index(idx_or_atomname)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname, int):
            idx = idx_or_atomname
        else:
            MolQueryError('Unrecognized atom name instance')
        # If empty, initialize

        self.atom_constraints[idx].append(constraint)

    def AppendBondConstraint(self,
                             idx_or_atomname1,
                             idx_or_atomname2,
                             bondconstraint):
        assert isinstance(bondconstraint, BondConstraint)
        if isinstance(idx_or_atomname1, str):
            try:
                idx1 = self.atom_names.index(idx_or_atomname1)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname1 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname1, int):
            idx1 = idx_or_atomname1
        else:
            MolQueryError('Unrecognized atom name instance')
        if isinstance(idx_or_atomname2, str):
            try:
                idx2 = self.atom_names.index(idx_or_atomname2)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname2 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname2, int):
            idx2 = idx_or_atomname2
        else:
            MolQueryError('Unrecognized atom name instance')
        # If empty, initialize

        self.bond_constraints.append([idx1, idx2, bondconstraint])

    def AppendDoubleBondStereoConstraint(self,
                                         idx_or_atomname1,
                                         idx_or_atomname2,
                                         idx_or_atomname3,
                                         idx_or_atomname4,
                                         doublebondstereoconstraint):

        assert isinstance(doublebondstereoconstraint,
                          DoubleBondStereoConstraint)
        if isinstance(idx_or_atomname1, str):
            try:
                idx1 = self.atom_names.index(idx_or_atomname1)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname1 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname1, int):
            idx1 = idx_or_atomname1
        else:
            MolQueryError('Unrecognized atom name instance')
        if isinstance(idx_or_atomname2, str):
            try:
                idx2 = self.atom_names.index(idx_or_atomname2)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname2 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname2, int):
            idx2 = idx_or_atomname2
        else:
            MolQueryError('Unrecognized atom name instance')
        if isinstance(idx_or_atomname3, str):
            try:
                idx3 = self.atom_names.index(idx_or_atomname3)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname3 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname3, int):
            idx3 = idx_or_atomname3
        else:
            MolQueryError('Unrecognized atom name instance')
        if isinstance(idx_or_atomname4, str):
            try:
                idx4 = self.atom_names.index(idx_or_atomname4)
            except Exception:
                s = "Undeclared atom name: '" + idx_or_atomname4 + "'"
                raise MolQueryError(s)
        elif isinstance(idx_or_atomname4, int):
            idx4 = idx_or_atomname4
        else:
            MolQueryError('Unrecognized atom name instance')
        # If empty, initialize

        self.double_bond_stereo_constraints.append([idx1, idx2, idx3, idx4,
                                                    doublebondstereoconstraint]
                                                   )

    def GetQueryMatches(self, mol, debug=0):
        mol = Chem.AddHs(mol)
        # test mol constraints
        for mol_constraint in self.mol_constraints:
            try:
                mol_constraint(mol)
            except Exception:
                return tuple()
        # Match mol substructure matches
        maxMatches = 10000
        rdkit_matches = mol.GetSubstructMatches(self.mol, uniquify=False,
                                                maxMatches=maxMatches)
        lenMatches = len(rdkit_matches)
        if lenMatches == maxMatches:
            print('\nMax RDKit substructure matches exceeded. All groups in',
                  'molecule may not have been determined.')
            print(Chem.MolToSmiles(Chem.RemoveHs(mol)))
        if debug:
            print('structure matches:' + str(rdkit_matches))
        # if no rdkit match
        if not rdkit_matches:
            return tuple()
        # Enhanced bond matching
        matches1 = list()
        for match_indice in rdkit_matches:
            try:
                for bond_constraint in self.bond_constraints:
                    idx1 = match_indice[bond_constraint[0]]
                    idx2 = match_indice[bond_constraint[1]]
                    bond_constraint[2](idx1, idx2, mol)
                matches1.append(match_indice)
            except Exception:
                continue
        # Enhanced atom matching
        matches2 = list()
        for match_indice in matches1:
            try:
                for i in range(0, len(match_indice)):
                    if i in self.atom_constraints:
                        # extract atom
                        # atom_index = match_indice[i]
                        atom = mol.GetAtomWithIdx(match_indice[i])
                        for atom_constraint in self.atom_constraints[i]:
                            atom_constraint(atom)
                matches2.append(match_indice)
            except Exception:
                continue
        # Enhanced bond stereo matching
        matches3 = list()
        for match_indice in matches2:
            try:
                for double_bond_stereo_constraint in\
                     self.double_bond_stereo_constraints:
                    idx1 = match_indice[double_bond_stereo_constraint[0]]
                    idx2 = match_indice[double_bond_stereo_constraint[1]]
                    idx3 = match_indice[double_bond_stereo_constraint[2]]
                    idx4 = match_indice[double_bond_stereo_constraint[3]]
                    double_bond_stereo_constraint[4](idx1, idx2,
                                                     idx3, idx4, mol)
                matches3.append(match_indice)
            except Exception:
                continue
        return tuple(matches3)
